import numpy as np
from vtool_ibeis import histogram as htool
from vtool_ibeis import keypoint as ktool
from vtool_ibeis import linalg as ltool
from vtool_ibeis import image as gtool
from vtool_ibeis import trig
import utool as ut
import ubelt as ub  # NOQA
from .util_math import TAU
try:
    import cv2
except ImportError:
    print('ERROR: import cv2 is failing!')
    cv2 = ut.DynStruct()
    cv2.INTER_LANCZOS4 = None
    cv2.INTER_CUBIC = None
    cv2.BORDER_CONSTANT = None
    cv2.BORDER_REPLICATE = None


def patch_gradient(patch, ksize=1, gaussian_weighted=False):
    patch_ = np.array(patch, dtype=np.float64)
    gradx = cv2.Sobel(patch_, cv2.CV_64F, 1, 0, ksize=ksize)
    grady = cv2.Sobel(patch_, cv2.CV_64F, 0, 1, ksize=ksize)
    if gaussian_weighted:
        gausspatch = gaussian_patch(shape=gradx.shape)
        gausspatch /= gausspatch.max()
        gradx *= gausspatch
        grady *= gausspatch
    return gradx, grady


def patch_mag(gradx, grady):
    return np.sqrt((gradx ** 2) + (grady ** 2))


def patch_ori(gradx, grady):
    """ returns patch orientation relative to the x-axis """
    gori = trig.atan2(grady, gradx)
    return gori


def get_test_patch(key='star', jitter=False):
    r"""
    Args:
        key (str):
        jitter (bool):

    Returns:
        ndarray: patch

    CommandLine:
        python -m vtool_ibeis.patch --test-get_test_patch --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> key = 'star2'
        >>> jitter = False
        >>> patch = get_test_patch(key, jitter)
        >>> pt.imshow(255 * patch)
        >>> pt.show_if_requested()
    """
    func = {
        'star2': get_star2_patch,
        'cross': get_cross_patch,
        'star': get_star_patch,
        'stripe': get_stripe_patch,
    }[key]
    patch = func(jitter)
    return patch


def make_test_image_keypoints(imgBGR, scale=1.0, skew=0, theta=0, shift=(0, 0)):
    h, w = imgBGR.shape[0:2]
    half_w, half_h = w / 2.0, h / 2.0
    x, y = (half_w - .5) + (w * shift[0]), (half_h - .5) + (h * shift[1])
    a = (half_w) * scale
    c = skew
    d = (half_h) * scale
    theta = theta
    kpts = np.array([[x, y, a, c, d, theta]], np.float32)
    return kpts


def get_no_symbol(variant='symbol', size=(100, 100)):
    r"""
    Returns:
        ndarray: errorimg

    CommandLine:
        python -m vtool_ibeis.patch --test-get_no_symbol --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> errorimg = get_no_symbol()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(errorimg)
        >>> ut.show_if_requested()
    """
    thickness = 2
    shape = (size[1], size[0], 3)
    errorimg = np.zeros(shape)
    center = (size[0] // 2, size[1] // 2)
    radius = min(center) - thickness
    color_bgr = [0, 0, 255]
    tau = 2 * np.pi
    angle = 45 / 360 * tau
    pt1 = (center[0] - int(np.sin(angle) * radius), center[1] - int(np.cos(angle) * radius))
    pt2 = (center[0] + int(np.sin(angle) * radius), center[1] + int(np.cos(angle) * radius))
    if variant == 'symbol':
        cv2.circle(errorimg, center, radius, color_bgr, thickness)
        cv2.line(errorimg, pt1, pt2, color_bgr, thickness)
    else:
        import vtool_ibeis as vt
        fontFace = cv2.FONT_HERSHEY_PLAIN
        org = (size[0] * .1, size[1] * .6)
        fontkw = dict(bottomLeftOrigin=False, fontScale=2.5, fontFace=fontFace)
        vt.draw_text(errorimg, 'NaN', org, thickness=2,
                     textcolor_rgb=color_bgr[::-1], **fontkw)
    return errorimg


def get_star_patch(jitter=False):
    """ test data patch """
    _, O = .1, .8
    patch = np.array([
        [_, _, _, O, _, _, _],
        [_, _, _, O, _, _, _],
        [_, _, O, O, O, _, _],
        [O, O, O, O, O, O, O],
        [_, O, O, O, O, O, _],
        [_, _, O, O, O, _, _],
        [_, O, O, O, O, O, _],
        [_, O, _, _, _, O, _],
        [O, _, _, _, _, _, O]])

    if jitter:
        patch += np.random.rand(*patch.shape) * .1
    return patch


def get_star2_patch(jitter=False):
    """ test data patch """
    _, i, O = .1, .8, .5
    patch = np.array([
        [_, _, _, _, _, _, _, O, O, _, _, _, _, _, _, _],
        [_, _, _, _, _, _, O, i, i, O, _, _, _, _, _, _],
        [_, _, _, _, _, _, O, i, i, O, _, _, _, _, _, _],
        [_, _, _, _, _, O, i, i, i, i, O, _, _, _, _, _],
        [O, O, O, O, O, O, i, i, i, i, O, O, O, O, O, O],
        [O, i, i, i, i, i, i, i, i, i, i, i, i, i, i, O],
        [_, O, i, i, i, i, O, i, i, O, i, i, i, i, O, _],
        [_, _, O, i, i, i, O, i, i, O, i, i, i, O, _, _],
        [_, _, _, O, i, i, O, i, i, O, i, i, O, _, _, _],
        [_, _, _, O, i, i, i, i, i, i, i, i, O, _, _, _],
        [_, _, O, i, i, i, i, i, i, i, i, i, i, O, _, _],
        [_, _, O, i, i, i, i, i, i, i, i, i, i, O, _, _],
        [_, O, i, i, i, i, i, O, O, i, i, i, i, i, O, _],
        [_, O, i, i, i, O, O, _, _, O, O, i, i, i, O, _],
        [O, i, i, O, O, _, _, _, _, _, _, O, O, i, i, O],
        [O, O, O, _, _, _, _, _, _, _, _, _, _, O, O, O]])

    if jitter:
        patch += np.random.rand(*patch.shape) * .1
    return patch


def get_cross_patch(jitter=False):
    """ test data patch """
    _, O = .1, .8
    patch = np.array([
        [_, _, O, O, O, _, _],
        [_, _, O, O, O, _, _],
        [_, _, O, O, O, _, _],
        [O, O, O, O, O, O, O],
        [O, O, O, O, O, O, O],
        [_, _, O, O, O, _, _],
        [_, _, O, O, O, _, _],
        [_, _, O, O, O, _, _],
        [_, _, O, O, O, _, _]])

    if jitter:
        patch += np.random.rand(*patch.shape) * .1
    return patch


def get_stripe_patch(jitter=False):
    """ test data patch """
    _, O = .1, .8
    patch = np.array([
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _],
        [O, O, O, _, _, _, _]])

    if jitter:
        patch += np.random.rand(*patch.shape) * .1
    return patch


def test_show_gaussian_patches2(shape=(19, 19)):
    r"""
    CommandLine:
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches2 --show
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches2 --show --shape=7,7
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches2 --show --shape=19,19
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches2 --show --shape=41,41
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches2 --show --shape=41,7

    References:
        http://matplotlib.org/examples/mplot3d/surface3d_demo.html

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> from mpl_toolkits.mplot3d import Axes3D  # NOQA
        >>> import plottool_ibeis as pt
        >>> shape = ut.get_argval(('--shape',), type_=list, default=[19, 19])
        >>> test_show_gaussian_patches2(shape=shape)
        >>> pt.show_if_requested()
    """
    from mpl_toolkits.mplot3d import Axes3D  # NOQA
    import plottool_ibeis as pt
    import numpy as np
    import matplotlib as mpl
    import vtool_ibeis as vt
    shape = tuple(map(int, shape))
    print('shape = %r' % (shape,))
    #shape = (27, 27)
    #shape = (7, 7)
    #shape = (41, 41)
    #shape = (5, 5)
    #shape = (3, 3)
    sigma_percent_list = [.1, .3, .5, .6, .7, .8, .9, .95, 1.0]
    #np.linspace(.1, 3, 9)
    ybasis = np.arange(shape[0])
    xbasis = np.arange(shape[1])
    xgrid, ygrid = np.meshgrid(xbasis, ybasis)
    fnum = pt.next_fnum()
    for sigma_percent in pt.param_plot_iterator(sigma_percent_list, fnum=fnum, projection='3d'):
        radius1 = shape[0]
        radius2 = shape[1]
        sigma1 = radius1 * sigma_percent
        sigma2 = radius2 * sigma_percent
        sigma = [sigma1, sigma2]
        gausspatch = vt.gaussian_patch(shape, sigma=sigma)
        #print(gausspatch)
        #pt.imshow(gausspatch * 255)
        pt.plot_surface3d(xgrid, ygrid, gausspatch, rstride=1, cstride=1,
                          cmap=mpl.cm.coolwarm, title='sigma_percent=%.3f' % (sigma_percent,))
    pt.update()
    pt.set_figtitle('2d gaussian kernels')


def show_gaussian_patch(shape, sigma1, sigma2):
    from mpl_toolkits.mplot3d import Axes3D  # NOQA
    import matplotlib as mpl
    import plottool_ibeis as pt
    import vtool_ibeis as vt
    ybasis = np.arange(shape[0])
    xbasis = np.arange(shape[1])
    xgrid, ygrid = np.meshgrid(xbasis, ybasis)
    sigma = [sigma1, sigma2]
    gausspatch = vt.gaussian_patch(shape, sigma=sigma)
    #print(gausspatch)
    #pt.imshow(gausspatch * 255)
    title = 'ksize=%r, sigma=%r' % (shape, (sigma1, sigma2),)
    pt.plot_surface3d(xgrid, ygrid, gausspatch, rstride=1, cstride=1,
                      cmap=mpl.cm.coolwarm, title=title)


def inverted_sift_patch(sift, dim=32):
    """
    Idea for inverted sift visualization

    CommandLine:
        python -m vtool_ibeis.patch test_sift_viz --show --name=star
        python -m vtool_ibeis.patch test_sift_viz --show --name=star2
        python -m vtool_ibeis.patch test_sift_viz --show --name=cross
        python -m vtool_ibeis.patch test_sift_viz --show --name=stripe

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> patch = vt.get_test_patch(ut.get_argval('--name', default='star'))
        >>> sift = vt.extract_feature_from_patch(patch)
        >>> siftimg = test_sift_viz(sift)
        >>> # Need to do some image blending
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.figure(fnum=1, pnum=(1, 2, 1))
        >>> pt.mpl_sift.draw_sift_on_patch(siftimg, sift)
        >>> pt.figure(fnum=1, pnum=(1, 2, 2))
        >>> patch2 = patch
        >>> patch2 = vt.rectify_to_uint8(patch2)
        >>> patch2 = vt.rectify_to_square(patch2)
        >>> pt.mpl_sift.draw_sift_on_patch(patch2, sift)
        >>> ut.show_if_requested()
    """
    import vtool_ibeis as vt
    # dim = 21
    pad = dim // 2 + (dim % 2)
    # pad = 0
    blocks = []
    for siftmags in ut.ichunks(sift, 8):
        thetas = np.linspace(0, TAU, 8, endpoint=False)
        # style = 'step'
        style = 'linear'
        block_parts = [gradient_fill(dim, theta, flip=0, style=style) * mag
                       for theta, mag in zip(thetas, siftmags)]
        block = np.add.reduce(block_parts)  # / sum(siftmags)
        # block = block[pad:-pad, pad:-pad]
        # block = gaussian_weight_patch(block, sigma=9)
        blocks.append(block)

    rows = []
    for row_blocks in ut.ichunks(blocks, 4):
        row = vt.stack_image_list(row_blocks, vert=False, overlap=pad)
        rows.append(row)
    siftimg = vt.stack_image_list(rows, vert=True, overlap=pad)
    siftimg /= siftimg.max()
    return siftimg


def gradient_fill(shape, theta=0, flip=False, vert=False, style='linear'):
    """
    FIXME: angle does not work properly

    CommandLine:
        python -m vtool_ibeis.patch gradient_fill --show

    Example:
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> shape = (9, 9)
        >>> #style = 'linear'
        >>> style = 'step'
        >>> theta = np.pi / 4
        >>> patch = vt.gradient_fill(shape, theta, style=style)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(vt.rectify_to_uint8(patch))
        >>> ut.show_if_requested()
    """
    if not isinstance(shape, tuple):
        shape = (shape, shape)
    import vtool_ibeis as vt
    patch = np.zeros(shape)
    if vert:
        vals = np.linspace(0, 1, shape[0])
    else:
        vals = np.linspace(0, 1, shape[1])

    if style == 'linear':
        vals = vals
    elif style == 'step':
        vals = vals > .5

    if flip:
        vals = vals[::-1]

    if vert:
        patch.T[:] = vals
    else:
        patch[:] = vals

    if theta != 0:
        patch = vt.rotate_image(patch, theta, interpolation='linear',
                                border_mode='replicate')

    patch = np.clip(patch, 0, 1)

    return patch


def test_show_gaussian_patches(shape=(19, 19)):
    r"""
    CommandLine:
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show --shape=7,7
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show --shape=17,17
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show --shape=41,41
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show --shape=29,29
        python -m vtool_ibeis.patch --test-test_show_gaussian_patches --show --shape=41,7

    References:
        http://matplotlib.org/examples/mplot3d/surface3d_demo.html

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> from mpl_toolkits.mplot3d import Axes3D  # NOQA
        >>> import plottool_ibeis as pt
        >>> shape = ut.get_argval(('--shape',), type_=list, default=[19, 19])
        >>> test_show_gaussian_patches(shape=shape)
        >>> pt.show_if_requested()
    """
    from mpl_toolkits.mplot3d import Axes3D  # NOQA
    import plottool_ibeis as pt
    import numpy as np
    import matplotlib as mpl
    import vtool_ibeis as vt
    shape = tuple(map(int, shape))
    print('shape = %r' % (shape,))
    #shape = (27, 27)
    #shape = (7, 7)
    #shape = (41, 41)
    #shape = (5, 5)
    #shape = (3, 3)
    sigma = 1.0
    sigma_list = [.1, .5, .825, .925, 1.0, 1.1, 1.2, 1.6, 2.0, 2.2, 3.0, 10.]
    #np.linspace(.1, 3, 9)
    ybasis = np.arange(shape[0])
    xbasis = np.arange(shape[1])
    xgrid, ygrid = np.meshgrid(xbasis, ybasis)
    fnum = pt.next_fnum()
    for sigma in pt.param_plot_iterator(sigma_list, fnum=fnum, projection='3d'):
        gausspatch = vt.gaussian_patch(shape, sigma=sigma)
        #print(gausspatch)
        #pt.imshow(gausspatch * 255)
        pt.plot_surface3d(xgrid, ygrid, gausspatch, rstride=1, cstride=1,
                          cmap=mpl.cm.coolwarm, title='sigma=%.3f' % (sigma,))
    pt.update()
    pt.set_figtitle('2d gaussian kernels')


def gaussian_patch(shape=(7, 7), sigma=1.0):
    """
    another version of the guassian_patch function. hopefully better

    References:
        http://docs.opencv.org/modules/imgproc/doc/filtering.html#getgaussiankernel

    Args:
        shape (tuple):  array dimensions
        sigma (float):

    CommandLine:
        python -m vtool_ibeis.patch --test-gaussian_patch --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> #shape = (7, 7)
        >>> shape = (24, 24)
        >>> sigma = None  # 1.0
        >>> gausspatch = gaussian_patch(shape, sigma)
        >>> sum_ = gausspatch.sum()
        >>> ut.assert_almost_eq(sum_, 1.0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(vt.norm01(gausspatch) * 255)
        >>> ut.show_if_requested()
    """
    if sigma is None:
        sigma = 0.3 * ((min(shape) - 1) * 0.5 - 1) + 0.8
    if isinstance(sigma, (float)):
        sigma1 = sigma2 = sigma
    else:
        sigma1, sigma2 = sigma
    # see hesaff/src/helpers.cpp : computeCircularGaussMask
    # HACK MAYBE: I think sigma is actually a sigma squared term?
    #sigma1 = np.sqrt(sigma1)
    #sigma2 = np.sqrt(sigma2)
    gauss_kernel_d0 = (cv2.getGaussianKernel(shape[0], sigma1))
    gauss_kernel_d1 = (cv2.getGaussianKernel(shape[1], sigma2))
    gausspatch = gauss_kernel_d0.dot(gauss_kernel_d1.T)
    return gausspatch


#@lru_cache(maxsize=1000)
#def gaussian_patch(width=3, height=3, shape=(7, 7), sigma=None, norm_01=True):
#    """
#    slow function that makes 2d gaussian image patch
#    It is essential that this function is cached!
#    """
#    # Build a list of x and y coordinates
#    half_width  = (width  / 2.0)
#    half_height = (height / 2.0)
#    gauss_xs = np.linspace(-half_width,  half_width,  shape[0])
#    gauss_ys = np.linspace(-half_height, half_height, shape[1])
#    # Iterate over the cartesian coordinate product and get pdf values
#    gauss_xys  = itertools.product(gauss_xs, gauss_ys)
#    gaussvals  = [ltool.gauss2d_pdf(x, y, sigma=sigma, mu=None)
#                  for (x, y) in gauss_xys]
#    # Reshape pdf values into a 2D image
#    gausspatch = np.array(gaussvals, dtype=np.float32).reshape(shape).T
#    if norm_01:
#        # normalize if requested
#        gausspatch -= gausspatch.min()
#        gausspatch /= gausspatch.max()
#    return gausspatch


def get_unwarped_patches(img, kpts):
    r"""
    Returns cropped unwarped (keypoint is still elliptical) patch around a
    keypoint

    Args:
        img (ndarray): array representing an image
        kpts (ndarrays): keypoint ndarrays in [x, y, a, c, d, theta] format
    Returns:
        tuple : (patches, subkpts) - the unnormalized patches from the img
            corresonding to the keypoint

    """
    _xs, _ys = ktool.get_xys(kpts)
    xyexnts = ktool.get_kpts_wh(kpts)
    patches = []
    subkpts = []

    for (kp, x, y, (sfx, sfy)) in zip(kpts, _xs, _ys, xyexnts):
        radius_x = sfx * 1.5
        radius_y = sfy * 1.5
        (chip_h, chip_w) = img.shape[0:2]
        # Get integer grid coordinates to crop at
        ix1, ix2, xm = htool.subbin_bounds(x, radius_x, 0, chip_w)
        iy1, iy2, ym = htool.subbin_bounds(y, radius_y, 0, chip_h)
        # Crop the keypoint out of the image
        patch = img[iy1:iy2, ix1:ix2]
        subkp = kp.copy()  # subkeypoint in patch coordinates
        subkp[0:2] = (xm, ym)
        patches.append(patch)
        subkpts.append(subkp)
    return patches, subkpts


def get_warped_patches(img, kpts, flags=cv2.INTER_LANCZOS4,
                       borderMode=cv2.BORDER_REPLICATE, patch_size=41,
                       use_cpp=False):
    r"""
    Returns warped (into a unit circle) patch around a keypoint

    FIXME:
        there is a slight translation difference in the way Python extracts
        patches and the way C++ extracts patches. C++ should be correct.
        TODO: have C++ able to extract color.

    Args:
        img (ndarray[uint8_t, ndim=2]): array representing an image
        kpts (ndarray[float32_t, ndim=2]): list of keypoint ndarrays in
            [[x, y, a, c, d, theta]] format
        flags (long): cv2 interpolation flags
        borderMode (long): cv2 border flags
        patch_size (int): resolution of resulting image patch

    Returns:
        (list, list) : (warped_patches, warped_subkpts) the normalized 41x41
            patches from the img corresonding to the keypoint

    CommandLine:
        python -m vtool_ibeis.patch --test-get_warped_patches --show --use_cpp
        python -m vtool_ibeis.patch --test-get_warped_patches --show --use_python

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> img_fpath = ut.grab_test_imgpath('carl')
        >>> img = vt.imread(img_fpath)
        >>> use_cpp = ut.get_argflag('--use_cpp')
        >>> kpts, desc = vt.extract_features(img_fpath)
        >>> kpts = kpts[0:1]
        >>> flags = cv2.INTER_LANCZOS4
        >>> borderMode = cv2.BORDER_REPLICATE
        >>> # execute function
        >>> (warped_patches, warped_subkpts) = get_warped_patches(img, kpts, flags, borderMode, use_cpp=use_cpp)
        >>> # verify results
        >>> print(np.array(warped_patches).shape)
        >>> print(ub.repr2(np.array(warped_subkpts), precision=2))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(warped_patches[0])
        >>> #pt.draw_kpts2(warped_subkpts, pts=True, rect=True)
        >>> pt.set_title('use_cpp = %r' % (use_cpp,))
        >>> pt.show_if_requested()
    """
    # TODO: CLEAN ME
    warped_patches = []
    warped_subkpts = []
    xs, ys = ktool.get_xys(kpts)
    # rotate relative to the gravity vector
    oris = ktool.get_oris(kpts)
    invV_mats = ktool.get_invV_mats(kpts, with_trans=False, ashomog=True)
    V_mats = ktool.invert_invV_mats(invV_mats)
    kpts_iter = zip(xs, ys, V_mats, oris)
    #patch_size = 41  # sf
    #cv2_warp_kwargs = {
    #    #'flags': cv2.INTER_LINEAR,
    #    #'flags': cv2.INTER_NEAREST,
    #    #'flags': cv2.INTER_LANCZOS4,
    #    'borderMode': cv2.BORDER_REPLICATE,
    #}
    #flags = cv2.INTER_CUBIC,
    #borderMode = cv2.BORDER_REPLICATE,
    if use_cpp:
        import pyhesaff
        warped_patches = pyhesaff.extract_patches(img, kpts)
        # FIXME:
        ss = np.sqrt(patch_size) * 3.0
        half_patch_size = patch_size / 2.0
        warped_subkpts.append(np.array((half_patch_size, half_patch_size, ss, 0., ss, 0)))
    else:
        for x, y, V, ori in kpts_iter:
            warped_patch, wkp = intern_warp_single_patch(img, x, y, ori, V,
                                                         patch_size,
                                                         flags=flags,
                                                         borderMode=borderMode)
            ## Build warped keypoints
            #wkp = np.array((patch_size / 2, patch_size / 2, ss, 0., ss, 0))
            warped_patches.append(warped_patch)
            warped_subkpts.append(wkp)
    return warped_patches, warped_subkpts


def intern_warp_single_patch(img, x, y, ori, V,
                             patch_size,
                             flags=cv2.INTER_CUBIC,
                             borderMode=cv2.BORDER_REPLICATE):
    r"""
    Sympy:
        # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
        from vtool_ibeis.patch import *  # NOQA
        import sympy
        from sympy.abc import theta
        ori = theta
        x, y, a, c, d, patch_size = sympy.symbols('x y a c d S')
        half_patch_size = patch_size / 2

        def sympy_rotation_mat3x3(radians):
            # TODO: handle array impouts
            sin_ = sympy.sin(radians)
            cos_ = sympy.cos(radians)
            R = np.array(((cos_, -sin_,  0),
                          (sin_,  cos_,  0),
                          (   0,     0,  1),))
            return sympy.Matrix(R)

        kpts = np.array([[x, y, a, c, d, ori]])
        kp = ktool.get_invV_mats(kpts, with_trans=True)[0]
        invV = sympy.Matrix(kp)
        V = invV.inv()
        ss = sympy.sqrt(patch_size) * 3.0
        T = sympy.Matrix(ltool.translation_mat3x3(-x, -y, None))  # Center the patch
        R = sympy_rotation_mat3x3(-ori)  # Rotate the centered unit circle patch
        S = sympy.Matrix(ltool.scale_mat3x3(ss, dtype=None))  # scale from unit circle to the patch size
        X = sympy.Matrix(ltool.translation_mat3x3(half_patch_size, half_patch_size, None))  # Translate back to patch-image coordinates

        sympy.MatMul(X, S, hold=True)

        def add_matmul_hold_prop(mat):
            #import functools
            def matmul_hold(other, hold=True):
                new = sympy.MatMul(mat, other, hold=hold)
                add_matmul_hold_prop(new)
                return new
            #matmul_hold = functools.partial(sympy.MatMul, mat, hold=True)
            setattr(mat, 'matmul_hold', matmul_hold)
        add_matmul_hold_prop(X)
        add_matmul_hold_prop(S)
        add_matmul_hold_prop(R)
        add_matmul_hold_prop(V)
        add_matmul_hold_prop(T)

        M = X.matmul_hold(S).matmul_hold(R).matmul_hold(V).matmul_hold(T)
        #M = X.multiply(S).multiply(R).multiply(V).multiply(T)


        V_full = R.multiply(V).multiply(T)
        sympy.latex(V_full)
        print(sympy.latex(R.multiply(V).multiply(T)))
        print(sympy.latex(X))
        print(sympy.latex(S))
        print(sympy.latex(R))
        print(sympy.latex(invV) + '^{-1}')
        print(sympy.latex(T))
    """
    cv2_warp_kwargs = {
        #'flags': cv2.INTER_LINEAR,
        #'flags': cv2.INTER_NEAREST,
        #'flags': cv2.INTER_CUBIC,
        #'flags': cv2.INTER_LANCZOS4,
        'flags': flags,
        #'borderMode': cv2.BORDER_REPLICATE,
        'borderMode': borderMode
    }
    #ut.embed()
    # FIXME: this works only because of add-hoc reasons.
    # need to more closely follow code in affine.cpp

    half_patch_size = patch_size / 2.0

    OLDWAY = True
    #OLDWAY = True
    if OLDWAY:
        ss = np.sqrt(patch_size) * 3.0
    else:
        # This seems to not work correctly
        mrSize = 3.0 * np.sqrt(3.0)
        sc = np.sqrt(1 / ktool.get_invVR_mats_sqrd_scale(V[None, :])[0])
        s = sc / mrSize
        mrScale = np.ceil(s * mrSize)
        patchImageSize = 2 * mrScale + 1
        imageToPatchScale = patchImageSize / patch_size
        ss = sc

    (h, w) = img.shape[0:2]
    # Translate to origin(0,0) = (x,y)
    T = ltool.translation_mat3x3(-x, -y)  # Center the patch
    # V - reshape and scale the centered patch to the unit circle
    R = ltool.rotation_mat3x3(-ori)  # Rotate the centered unit circle patch
    S = ltool.scale_mat3x3(ss)  # scale from unit circle to the patch size
    X = ltool.translation_mat3x3(half_patch_size, half_patch_size)  # Translate back to patch-image coordinates
    M = X.dot(S).dot(R).dot(V).dot(T)
    # Prepare to warp
    dsize = np.ceil([patch_size, patch_size]).astype(int)
    # Warp
    #warped_patch = gtool.warpAffine(img, M, dsize)
    warped_patch = cv2.warpAffine(img, M[0:2], tuple(dsize), **cv2_warp_kwargs)
    # Build warped keypoints
    wkp = np.array((half_patch_size, half_patch_size, ss, 0., ss, 0))

    if not OLDWAY:
        # FIXME: this is still not what is done in affine.cpp
        if imageToPatchScale > 0.4:
            #ksize = (patchImageSize / 2, patchImageSize / 2)
            #sigmaX, sigmaY = (patchImageSize / 2, patchImageSize / 2)
            #ut.embed()
            sigma = imageToPatchScale * 1.5
            GaussianBlurInplace(warped_patch, sigma)
    else:
        sigma = 1.5
        GaussianBlurInplace(warped_patch, sigma)
    return warped_patch, wkp


def generate_to_patch_transforms(kpts, patch_size=41):
    import vtool_ibeis as vt
    xs, ys = vt.get_xys(kpts)
    # rotate relative to the gravity vector
    oris = vt.get_oris(kpts)
    invV_mats = vt.get_invV_mats(kpts, with_trans=False, ashomog=True)
    V_mats = vt.invert_invV_mats(invV_mats)
    kpts_iter = zip(xs, ys, V_mats, oris)
    # HACK, this is using the old method
    half_patch_size = patch_size / 2.0
    ss = np.sqrt(patch_size) * 3.0

    for x, y, V, ori in kpts_iter:
        T = vt.translation_mat3x3(-x, -y)  # Center the patch
        # V - reshape and scale the centered patch to the unit circle
        R = vt.rotation_mat3x3(-ori)  # Rotate the centered unit circle patch
        S = vt.scale_mat3x3(ss)  # scale from unit circle to the patch size
        X = vt.translation_mat3x3(half_patch_size, half_patch_size)  # Translate back to patch-image coordinates
        M = X.dot(S).dot(R).dot(V).dot(T)
        yield M


def patch_gaussian_weighted_average_intensities(probchip, kpts_):
    """
    """
    import vtool_ibeis as vt
    patch_size = 41
    M_iter = vt.generate_to_patch_transforms(kpts_, patch_size)
    dsize = np.ceil([patch_size, patch_size]).astype(int)
    # Preallocate patch
    patch = np.empty(dsize[::-1], dtype=np.uint8)
    weighted_patch = np.empty(dsize[::-1], dtype=np.float64)
    weight_list = []
    sigma = 0.3 * ((min(patch.shape[0:1]) - 1) * 0.5 - 1) + 0.8
    gauss_kernel_d0 = (cv2.getGaussianKernel(patch.shape[0], sigma))
    gauss_kernel_d1 = (cv2.getGaussianKernel(patch.shape[1], sigma)).T
    warpkw = dict(flags=cv2.INTER_LANCZOS4, borderMode=cv2.BORDER_REPLICATE)
    for M in M_iter:
        patch = cv2.warpAffine(probchip, M[0:2], tuple(dsize), dst=patch, **warpkw)
        vt.GaussianBlurInplace(patch, 1.5)
        weighted_patch = np.divide(patch, 255., out=weighted_patch)
        weighted_patch = np.multiply(weighted_patch, gauss_kernel_d0, out=weighted_patch)
        weighted_patch = np.multiply(weighted_patch, gauss_kernel_d1, out=weighted_patch)
        weight = weighted_patch.sum()
        weight_list.append(weight)
    return weight_list


def gaussian_average_patch(patch, sigma=None, copy=True):
    """

    Args:
        patch (ndarray):
        sigma (float):

    CommandLine:
        python -m vtool_ibeis.patch --test-gaussian_average_patch

    References:
        http://docs.opencv.org/modules/imgproc/doc/filtering.html#getgaussiankernel

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> patch = get_star_patch()
        >>> #sigma = 1.6
        >>> sigma = None
        >>> result = gaussian_average_patch(patch, sigma)
        >>> print(result)
        0.414210641527

    Ignore:
        import utool as ut
        import plottool_ibeis as pt
        import vtool_ibeis as vt
        import cv2
        gauss_kernel_d0 = (cv2.getGaussianKernel(patch.shape[0], sigma))
        gauss_kernel_d1 = (cv2.getGaussianKernel(patch.shape[1], sigma))
        weighted_patch = patch.copy()
        weighted_patch = np.multiply(weighted_patch,   gauss_kernel_d0)
        weighted_patch = np.multiply(weighted_patch.T, gauss_kernel_d1).T
        gaussian_kern2 = gauss_kernel_d0.dot(gauss_kernel_d1.T)
        fig = pt.figure(fnum=1, pnum=(1, 3, 1), doclf=True, docla=True)
        pt.imshow(patch * 255)
        fig = pt.figure(fnum=1, pnum=(1, 3, 2))
        pt.imshow(ut.norm_zero_one(gaussian_kern2) * 255.0)
        fig = pt.figure(fnum=1, pnum=(1, 3, 3))
        pt.imshow(ut.norm_zero_one(weighted_patch) * 255.0)
        pt.update()
    """
    if sigma is None:
        sigma = 0.3 * ((min(patch.shape[0:1]) - 1) * 0.5 - 1) + 0.8
    gauss_kernel_d0 = (cv2.getGaussianKernel(patch.shape[0], sigma))
    gauss_kernel_d1 = (cv2.getGaussianKernel(patch.shape[1], sigma))
    # assert gauss_kernel_d0.dot(gauss_kernel_d1.T).sum() == 1
    if copy:
        weighted_patch = patch.copy()
    else:
        weighted_patch = patch
    weighted_patch = np.multiply(weighted_patch,   gauss_kernel_d0)
    weighted_patch = np.multiply(weighted_patch.T, gauss_kernel_d1).T
    # TODO: use new guass patch weighting function here
    average = weighted_patch.sum()
    return average


def gaussian_weight_patch(patch, sigma=None):
    """
    Applies two one dimensional gaussian operations to a patch which
    effectively weights it by a 2-dimensional gaussian. This is efficient
    because the actually 2-d gaussian never needs to be allocated.

    test_show_gaussian_patches
    """
    #patch = np.ones(patch.shape)
    if sigma is None:
        sigma0 = (patch.shape[0] / 2) * .95
        sigma1 = (patch.shape[1] / 2) * .95
    else:
        sigma0 = sigma1 = sigma
    #sigma0 = (patch.shape[0] / 2) * .5
    #sigma1 = (patch.shape[1] / 2) * .5
    gauss_kernel_d0 = (cv2.getGaussianKernel(patch.shape[0], sigma0))
    gauss_kernel_d1 = (cv2.getGaussianKernel(patch.shape[1], sigma1))
    gauss_kernel_d0 /= gauss_kernel_d0.max()
    gauss_kernel_d1 /= gauss_kernel_d1.max()
    weighted_patch = (patch * gauss_kernel_d0.T) * gauss_kernel_d1
    return weighted_patch


def get_warped_patch(imgBGR, kp, gray=False, flags=cv2.INTER_LANCZOS4,
                     borderMode=cv2.BORDER_REPLICATE, patch_size=41):
    r"""
    Returns warped (into a unit circle) patch around a keypoint

    Args:
        img (ndarray): array representing an image
        kpt (ndarray): keypoint ndarray in [x, y, a, c, d, theta] format

    Returns:
        (ndarray, ndarray) : (wpatch, wkp) the normalized 41x41 patches from
            the img corresonding to the keypoint
    """
    kpts = np.array([kp])
    wpatches, wkpts = get_warped_patches(imgBGR, kpts, flags=flags,
                                         borderMode=borderMode,
                                         patch_size=patch_size)
    wpatch = wpatches[0]
    wkp = wkpts[0]
    if gray and len(wpatch.shape) > 2:
        wpatch = gtool.cvt_BGR2L(wpatch)
    return wpatch, wkp


def GaussianBlurInplace(img, sigma, size=None):
    """
    simulates code from helpers.cpp in hesaff

    Args:
        img (ndarray):
        sigma (flaot):

    CommandLine:
        python -m vtool_ibeis.patch --test-GaussianBlurInplace:0 --show
        python -m vtool_ibeis.patch --test-GaussianBlurInplace:1 --show


    References;
        http://www.cse.usf.edu/~r1k/MachineVisionBook/MachineVision.files/MachineVision_Chapter4.pdf
        http://en.wikipedia.org/wiki/Scale_space_implementation
        http://www.cse.psu.edu/~rtc12/CSE486/lecture10_6pp.pdf

    Notes:
        The product of the convolution of two Gaussian functions with spread
        sigma is a Gaussian function with spread sqrt(2)*sigma scaled by the
        area of the Gaussian filter

    Example0:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> from mpl_toolkits.mplot3d import Axes3D  # NOQA
        >>> import plottool_ibeis as pt
        >>> img = get_test_patch('star2')
        >>> img_orig = img.copy()
        >>> sigma = .8
        >>> GaussianBlurInplace(img, sigma)
        >>> fig = pt.figure(fnum=1, pnum=(1, 3, 1))
        >>> size = int((2.0 * 3.0 * sigma + 1.0))
        >>> if not size & 1:  # check if even
        >>>     size += 1
        >>> ksize = (size, size)
        >>> fig.add_subplot(1, 3, 1, projection='3d')
        >>> show_gaussian_patch(ksize, sigma, sigma)
        >>> pt.imshow(img_orig * 255, fnum=1, pnum=(1, 3, 2))
        >>> pt.imshow(img * 255, fnum=1, pnum=(1, 3, 3))
        >>> pt.show_if_requested()

    Example1:
        >>> # DISABLE_DOCTEST
        >>> # demonstrate cascading smoothing property
        >>> # THIS ISNT WORKING WHY???
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> from mpl_toolkits.mplot3d import Axes3D  # NOQA
        >>> import plottool_ibeis as pt
        >>> img = get_test_patch('star2')
        >>> img1 = img.copy()
        >>> img2 = img.copy()
        >>> img3 = img.copy()
        >>> img4 = img.copy()
        >>> img_orig = img.copy()
        >>> sigma1 = .6
        >>> sigma2 = .9
        >>> sigma3 = sigma1 + sigma2
        >>> size = 7
        >>> # components
        >>> GaussianBlurInplace(img1, sigma1, size)
        >>> GaussianBlurInplace(img2, sigma2, size)
        >>> # all in one shot
        >>> GaussianBlurInplace(img3, sigma3, size)
        >>> # addative
        >>> GaussianBlurInplace(img4, sigma1, size)
        >>> GaussianBlurInplace(img4, sigma2, size)
        >>> print((img4 - img3).sum())
        >>> # xdoctest: +REQUIRES(--show)
        >>> fig = pt.figure(fnum=1, pnum=(2, 4, 1))
        >>> ksize = (size, size)
        >>> #fig.add_subplot(1, 3, 1, projection='3d')
        >>> fig.add_subplot(2, 4, 1, projection='3d')
        >>> show_gaussian_patch(ksize, sigma1, sigma1)
        >>> fig.add_subplot(2, 4, 2, projection='3d')
        >>> show_gaussian_patch(ksize, sigma2, sigma2)
        >>> fig.add_subplot(2, 4, 3, projection='3d')
        >>> show_gaussian_patch(ksize, sigma3, sigma3)
        >>> pt.imshow(img_orig * 255, fnum=1, pnum=(2, 4, 4))
        >>> pt.imshow(img1 * 255, fnum=1, pnum=(2, 4, 5), title='%r' % (sigma1))
        >>> pt.imshow(img2 * 255, fnum=1, pnum=(2, 4, 6), title='%r' % (sigma2))
        >>> pt.imshow(img3 * 255, fnum=1, pnum=(2, 4, 7), title='%r' % (sigma3))
        >>> pt.imshow(img4 * 255, fnum=1, pnum=(2, 4, 8), title='%r + %r' % (sigma1, sigma2))
        >>> pt.show_if_requested()
    """
    if size is None:
        size = int((2.0 * 3.0 * sigma + 1.0))
        if not size & 1:  # check if even
            size += 1
    else:
        assert size & 1, 'size must be odd'
    ksize = (size, size)
    cv2.GaussianBlur(img, ksize, sigmaX=sigma, sigmaY=sigma, dst=img, borderType=cv2.BORDER_REPLICATE)
    return img


def get_unwarped_patch(imgBGR, kp, gray=False):
    """Returns unwarped warped patch around a keypoint

    Args:
        img (ndarray): array representing an image
        kpt (ndarray): keypoint ndarray in [x, y, a, c, d, theta] format
    Returns:
        tuple : (wpatch, wkp) the normalized 41x41 patches from the img corresonding to the keypoint
    """
    kpts = np.array([kp])
    upatches, ukpts = get_unwarped_patches(imgBGR, kpts)
    upatch = upatches[0]
    ukp = ukpts[0]
    if gray:
        upatch = gtool.cvt_BGR2L(upatch)
    return upatch, ukp


def find_kpts_direction(imgBGR, kpts, DEBUG_ROTINVAR=False):
    r"""
    Args:
        imgBGR (ndarray[uint8_t, ndim=2]):  image data in opencv format (blue, green, red)
        kpts (ndarray[float32_t, ndim=2]):  keypoints

    Returns:
        ndarray[float32_t, ndim=2]: kpts -  keypoints

    CommandLine:
        python -m vtool_ibeis.patch --test-find_kpts_direction
    """

    ori_list = []
    #gravity_ori = ktool.GRAVITY_THETA
    for kp in kpts:
        new_oris = find_dominant_kp_orientations(imgBGR, kp, DEBUG_ROTINVAR=DEBUG_ROTINVAR)
        # FIXME USE MULTIPLE ORIENTATIONS
        ori = new_oris[0]
        ori_list.append(ori)
    _oris = np.array(ori_list, dtype=kpts.dtype)
    #_oris -= gravity_ori  % TAU  # normalize w.r.t. gravity
    # discard old orientation if they exist
    kpts2 = np.vstack((kpts[:, 0:5].T, _oris)).T
    return kpts2


def draw_kp_ori_steps():
    """
    Shows steps in orientation estimation

    CommandLine:
        python -m vtool_ibeis.patch --test-draw_kp_ori_steps --show --fname=zebra.png --fx=121
        python -m vtool_ibeis.patch --test-draw_kp_ori_steps --show --interact
        python -m vtool_ibeis.patch --test-draw_kp_ori_steps --save ~/latex/crall-candidacy-2015/figures/test_fint_kp_direction.jpg --dpath figures '--caption=visualization of the steps in the computation of the dominant gradient orientations.' --figsize=14,9 --dpi=160 --height=2.65  --left=.04 --right=.96 --top=.95 --bottom=.05 --wspace=.1 --hspace=.1

        python -m vtool_ibeis.patch --test-draw_kp_ori_steps --dpath ~/latex/crall-candidacy-2015/ --save figures/draw_kp_ori_steps.jpg  --figsize=14,9 --dpi=180 --height=2.65 --left=.04 --right=.96 --top=.95 --bottom=.05 --wspace=.1 --hspace=.1 --diskshow

        python -m vtool_ibeis.patch --test-draw_kp_ori_steps --dpath ~/latex/crall-candidacy-2015/ --save figures/draw_kp_ori_steps.jpg  --figsize=14,9 --dpi=180  --djust=.04,.05,.1 --diskshow --fname=zebra.png --fx=121

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> import plottool_ibeis as pt
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> draw_kp_ori_steps()
        >>> pt.show_if_requested()
    """
    # build test data
    import utool as ut
    import plottool_ibeis as pt

    if True:
        from ibeis.scripts.thesis import TMP_RC
        import matplotlib as mpl
        mpl.rcParams.update(TMP_RC)
    np.random.seed(0)
    USE_COMMANLINE = True
    if USE_COMMANLINE:
        kpts, vecs, imgBGR = pt.viz_keypoints.testdata_kpts()
        fx = ut.get_argval('--fx', type_=int, default=0)
        kp = kpts[fx]
    else:
        fx = 0
        #imgBGR = get_test_patch('stripe', jitter=True)
        #imgBGR = get_test_patch('star', jitter=True)
        imgBGR = get_test_patch('star2', jitter=True)
        #imgBGR = get_test_patch('cross', jitter=False)
        #imgBGR = cv2.resize(imgBGR, (41, 41), interpolation=cv2.INTER_LANCZOS4)
        imgBGR = cv2.resize(imgBGR, (41, 41), interpolation=cv2.INTER_CUBIC)
        theta = 0  # 3.4  # TAU / 16
        #kpts = make_test_image_keypoints(imgBGR, scale=.9, theta=theta)
        kpts = make_test_image_keypoints(imgBGR, scale=.3, theta=theta, shift=(.3, .1))
        kp = kpts[0]
    bins = 36
    maxima_thresh = .8
    converge_lists = []

    def exec_internals_find_patch_dominant_orientations(patch, bins, maxima_thresh, old_ori):
        # TODO: can use ut.exec_func_src instead
        # <HACKISH>
        # exec source code from find_patch_dominant_orientations to steal its
        # local variables
        # ARGS: patch, bins=36, maxima_thresh=.8, DEBUG_ROTINVAR

        DEBUG_ROTINVAR = False
        globals_ = globals()
        locals_ = locals()
        keys = 'patch, gradx, grady, gmag, gori, hist, centers, gori_weights'.split(', ')
        internal_tup = ut.exec_func_src(find_patch_dominant_orientations, globals_, locals_, key_list=keys, update=True)
        submax_ori_offsets = globals_['submax_ori_offsets']
        new_oris = (old_ori + (submax_ori_offsets - ktool.GRAVITY_THETA)) % TAU
        return new_oris, internal_tup
        # </HACKISH>

    INTERACTIVE_ITERATION = ut.get_argflag('--interact')

    while True:
        patch, wkp = get_warped_patch(imgBGR, kp, gray=True,
                                      #flags=cv2.INTER_LANCZOS4,
                                      flags=cv2.INTER_CUBIC,
                                      borderMode=cv2.BORDER_CONSTANT)
        old_ori = kp[-1]
        # Execute test function
        new_oris, internal_tup = exec_internals_find_patch_dominant_orientations(patch, bins, maxima_thresh, old_ori)
        patch, gradx, grady, gmag, gori, hist, centers, gori_weights = internal_tup
        if INTERACTIVE_ITERATION:
            # Change rotation
            print('new_oris = %r' % (new_oris,))
            print('bins = %r' % (bins,))
            for count, ori in enumerate(new_oris):
                if count >= len(converge_lists):
                    converge_lists.append([])
                converge_lists[count].append(ori)

            # Show any new keypoints that were created
            kpts = [kp.copy() for ori in new_oris]
            for count, ori in enumerate(new_oris):
                kpts[count][-1] = ori
            kpts = np.array(kpts)

            kp[-1] = new_oris[0]

            show_patch_orientation_estimation(imgBGR, kpts, patch, gradx, grady, gmag, gori, hist, centers, gori_weights, fx=fx)
            pt.figure(fnum=2, doclf=True)
            colors = pt.distinct_colors(len(converge_lists))
            print(len(converge_lists))
            for color, converge_list in zip(colors, converge_lists):
                pt.plot(converge_list, 'o-', color=color)
            pt.gca().set_ylim(0, TAU)
            pt.dark_background()
            pt.present()
            #pt.update()
            input('next')
        else:
            show_patch_orientation_estimation(imgBGR, kpts, patch, gradx, grady, gmag, gori, hist, centers, gori_weights, fx=fx)
            # pt.present()
            print('no interaction')
            break


def show_patch_orientation_estimation(imgBGR, kpts, patch, gradx, grady, gmag,
                                      gori, hist, centers, gori_weights,
                                      fx=None):
    import plottool_ibeis as pt
    import vtool_ibeis as vt
    # DRAW TEST INFO
    fnum = 1
    pt.figure(fnum=1, doclf=True, docla=True)
    gorimag = pt.color_orimag(gori, gmag=gmag, gmag_is_01=False)
    nRows, nCols = pt.get_square_row_cols(8)
    nRows += 1
    pnum_ = pt.make_pnum_nextgen(nRows, nCols)
    # hack
    imgBGR_ = imgBGR if imgBGR.max() > 1 else imgBGR * 255
    patch_ = patch if patch.max() > 1 else patch * 255
    pt.imshow(imgBGR_, update=True, fnum=fnum, pnum=pnum_(), title='input image')
    colors = pt.distinct_colors(len(kpts))
    if fx is None:
        pt.draw_kpts2(kpts, rect=True, ori=True, ell_color=colors)
    else:
        pt.draw_kpts2(kpts[fx:fx + 1], rect=True, ori=True, colors=[pt.ORANGE])
    pt.imshow(patch_, fnum=fnum, pnum=pnum_(), title='sampled patch')
    def normalize_grad_img(grad_):
        #return np.abs(grad_) * 255
        return vt.norm01(np.abs(grad_)) * 255
        #return vt.norm01(grad_) * 255
    pt.imshow(normalize_grad_img(gradx ** 2), fnum=fnum, pnum=pnum_(), title='gradx ** 2')
    pt.imshow(normalize_grad_img(grady ** 2), fnum=fnum, pnum=pnum_(), title='grady ** 2')
    pt.imshow(normalize_grad_img(gmag), fnum=fnum, pnum=pnum_(), title='mag')
    pt.imshow(normalize_grad_img(gori_weights), fnum=fnum, pnum=pnum_(), title='weighted mag')
    #pt.imshow(ut.norm_zero_one(gori) * 255, fnum=fnum, pnum=pnum_(), title='ori')
    stride = ut.get_argval('--stride', default=1)
    pt.draw_vector_field(gradx, grady, stride=stride, pnum=pnum_(), fnum=fnum,
                         title='gori (vec)')
    pt.imshow(gorimag, fnum=fnum, pnum=pnum_(), title='ori-color')
    if not ut.get_argflag('--noweighted-gori'):
        pt.color_orimag_colorbar(gori * gori_weights)
    else:
        pt.color_orimag_colorbar(gori)
    pt.figure(fnum=fnum, pnum=(nRows, 1, nRows))
    bin_colors = pt.get_orientation_color(centers)
    pt.draw_hist_subbin_maxima(hist, centers, bin_colors=bin_colors,
                               maxima_thresh=.8)
    ax = pt.gca()
    ax.set_xlabel('radians')
    ax.set_ylabel('weight')


def test_ondisk_find_patch_fpath_dominant_orientations(patch_fpath, bins=36,
                                                       maxima_thresh=.8,
                                                       DEBUG_ROTINVAR=True):
    r"""
    Args:
        patch_fpath (?):
        bins (int):
        maxima_thresh (float):

    CommandLine:
        python -m vtool_ibeis.patch --test-test_ondisk_find_patch_fpath_dominant_orientations

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> # build test data
        >>> patch_fpath = ut.get_argval('--patch-fpath', type_=str, default=ut.grab_test_imgpath('star.png'))
        >>> bins = 36
        >>> maxima_thresh = 0.8
        >>> test_ondisk_find_patch_fpath_dominant_orientations(patch_fpath, bins, maxima_thresh)
        >>> pt.show_if_requested()
    """
    import vtool_ibeis as vt
    patch = vt.imread(patch_fpath, grayscale=True)
    #submax_ori = submaxima_x[submaxima_y.argmax()]
    #ori_offsets = [submax_ori]  # normalize w.r.t. gravity
    return find_patch_dominant_orientations(patch, bins=bins, maxima_thresh=maxima_thresh, DEBUG_ROTINVAR=DEBUG_ROTINVAR)


def find_patch_dominant_orientations(patch, bins=36, maxima_thresh=.8,
                                     DEBUG_ROTINVAR=False):
    """
    helper
    """
    import plottool_ibeis as pt
    gradx, grady = patch_gradient(patch, gaussian_weighted=False)
    gori = patch_ori(gradx, grady)
    gmag = patch_mag(gradx, grady)
    gaussian_weighted = True
    # do gaussian weighting correctly
    gori_weights = gaussian_weight_patch(gmag) if gaussian_weighted else gmag
    if DEBUG_ROTINVAR:
        print(ktool.kpts_docrepr(patch[::10, ::10], 'PATCH[::10]', False))
        print(ktool.kpts_docrepr(gradx[::10, ::10], 'gradx[::10]', False))
        print(ktool.kpts_docrepr(grady[::10, ::10], 'grady[::10]', False))
        #print(ktool.kpts_docrepr(patch, 'PATCH', False))
        print(ktool.kpts_docrepr(gori[::10, ::10], 'ORI[::10]', False))
        print(ktool.kpts_docrepr(gmag[::10, ::10], 'MAG[::10]', False))
        print(ktool.kpts_docrepr(gori_weights[::10, ::10], 'WEIGHTS[::10]', False))
    # FIXME: Not taking account to gmag
    #bins = 3
    #bins = 8
    hist, centers = get_orientation_histogram(gori, gori_weights, bins=bins)
    # Find submaxima
    submaxima_x, submaxima_y = htool.argsubmaxima(hist, centers,
                                                  maxima_thresh=maxima_thresh)
    if DEBUG_ROTINVAR:
        htool.show_ori_image(gori, gori_weights / 255.0, patch, gradx, grady)
        pt.set_figtitle('python orimg')
    submax_ori_offsets = submaxima_x
    if DEBUG_ROTINVAR:
        print('submaxima_x, submaxima_y = %r, %r' % (submaxima_x, submaxima_y,))
        htool.show_hist_submaxima(hist, centers=centers)
        pt.set_figtitle('python hist')
        pt.df2.plt.show()
    return submax_ori_offsets


def testdata_patch():
    import plottool_ibeis as pt
    kpts, vecs, imgBGR = pt.viz_keypoints.testdata_kpts()
    fx = ut.get_argval('--fx', type_=int, default=0)
    kp = kpts[fx]
    patch, wkp = get_warped_patch(imgBGR, kp, gray=True,
                                  #flags=cv2.INTER_LANCZOS4,
                                  flags=cv2.INTER_CUBIC,
                                  borderMode=cv2.BORDER_CONSTANT)
    return patch


def find_dominant_kp_orientations(imgBGR, kp, bins=36, maxima_thresh=.8,
                                  DEBUG_ROTINVAR=False):
    r"""

    References:
        http://szeliski.org/Book/drafts/SzeliskiBook_20100903_draft.pdf
        page 219

        http://www.cs.berkeley.edu/~malik/cs294/lowe-ijcv04.pdf
        page 13.

        Lowe uses a 36-bin histogram of edge orientations weigted by a gaussian
        distance to the center and gradient magintude. He finds all peaks within
        80% of the global maximum. Then he fine tunes the orientation using a
        3-binned parabolic fit. Multiple orientations (and hence multiple
        keypoints) can be returned, but empirically only about 15% will have
        these and they do tend to be important.

    Returns:
        float: ori_offset - offset of current orientation to dominant orientation

    CommandLine:
        python -m vtool_ibeis.patch --test-find_dominant_kp_orientations

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> np.random.seed(0)
        >>> #imgBGR = get_test_patch('cross', jitter=False)
        >>> img_fpath = ut.grab_test_imgpath('star.png')
        >>> imgBGR = vt.imread(img_fpath)
        >>> kpts, vecs = vt.extract_features(img_fpath)
        >>> assert len(kpts) == 1
        >>> kp = kpts[0]
        >>> print('kp = \n' + (vt.kp_cpp_infostr(kp)))
        >>> bins = 36
        >>> maxima_thresh = .8
        >>> # execute function
        >>> new_oris = find_dominant_kp_orientations(imgBGR, kp, bins,
        >>>                                          maxima_thresh,
        >>>                                          DEBUG_ROTINVAR=True)
        >>> # verify results
        >>> result = 'new_oris = %r' % (new_oris,)
    """
    patch, wkp = get_warped_patch(imgBGR, kp, gray=True,
                                  #flags=cv2.INTER_LANCZOS4,
                                  flags=cv2.INTER_CUBIC,
                                  borderMode=cv2.BORDER_CONSTANT)

    if DEBUG_ROTINVAR:
        pass
        if True:
            import plottool_ibeis as pt
            fnum = pt.next_fnum()
            pt.imshow(imgBGR, fnum=fnum, pnum=(1, 2, 1))
            pt.draw_kpts2(np.array([kp]), pts=True, rect=True)
            pt.imshow(patch, fnum=fnum, pnum=(1, 2, 2))
            pt.draw_kpts2(np.array([wkp]), pts=True, rect=True)
    # Compute new orientation(s) for this keypoint
    submax_ori_offsets = find_patch_dominant_orientations(patch, bins=bins,
                                                          maxima_thresh=maxima_thresh,
                                                          DEBUG_ROTINVAR=DEBUG_ROTINVAR)
    old_ori = kp[-1]
    new_oris = (old_ori + (submax_ori_offsets + ktool.GRAVITY_THETA)) % TAU
    return new_oris


def get_orientation_histogram(gori, gori_weights, bins=36, DEBUG_ROTINVAR=False):
    r"""
    Args:
        gori (?):
        gori_weights (?):
        bins (int):

    Returns:
        tuple: (hist, centers)

    CommandLine:
        python -m vtool_ibeis.patch --test-get_orientation_histogram

    Ignore:
        print(vt.kpts_docrepr(gori, 'gori = '))
        print(vt.kpts_docrepr(gori_weights, 'gori_weights = '))

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.patch import *  # NOQA
        >>> # build test data
        >>> gori = np.array([[ 0.  ,  0.  ,  3.14,  3.14,  0.  ],
        ...                  [ 4.71,  6.15,  3.13,  3.24,  4.71],
        ...                  [ 4.71,  4.61,  0.5 ,  4.85,  4.71],
        ...                  [ 1.57,  6.28,  3.14,  3.14,  1.57],
        ...                  [ 0.  ,  0.  ,  3.14,  3.14,  0.  ]])
        >>> gori_weights = np.array([[ 0.  ,  0.11,  0.02,  0.13,  0.  ],
        ...                          [ 0.02,  0.19,  0.02,  0.21,  0.02],
        ...                          [ 0.11,  0.16,  0.  ,  0.13,  0.11],
        ...                          [ 0.  ,  0.17,  0.02,  0.19,  0.  ],
        ...                          [ 0.  ,  0.11,  0.02,  0.13,  0.  ]])
        >>> bins = 36
        >>> # execute function
        >>> (hist, centers) = get_orientation_histogram(gori, gori_weights, bins)
        >>> # verify results
        >>> result = str((hist, centers))
        >>> print(result)
    """
    # Get wrapped histogram (because we are finding a direction)
    flat_oris = gori.flatten()
    flat_weights = gori_weights.flatten()
    range_ = (0, TAU)
    # FIXME: this does not do linear interpolation
    #hist_, edges_ = np.histogram(flat_oris, range=range_, bins=bins, weights=flat_weights)
    # Compute histogram where orientations split weight between bins
    hist_, edges_ = htool.interpolated_histogram(flat_oris, flat_weights,
                                                 range_, bins,
                                                 interpolation_wrap=True,
                                                 _debug=DEBUG_ROTINVAR)
    # Duplicate the first and last edges so neighbor information is contiguous
    hist, edges = htool.wrap_histogram(hist_, edges_, _debug=DEBUG_ROTINVAR)
    centers = htool.hist_edges_to_centers(edges)
    return hist, centers


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m vtool_ibeis.patch
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Wrapper around flann (with caching)

python -c "import vtool_ibeis, doctest; print(doctest.testmod(vtool_ibeis.nearest_neighbors))"
"""
from os.path import exists, normpath, join
import utool as ut
import ubelt as ub
import numpy as np

try:
    from vtool_ibeis._pyflann_backend import FLANN_CLS
    from vtool_ibeis._pyflann_backend import pyflann
except Exception:
    FLANN_CLS = None
    pyflann = None


class AnnoyWrapper(object):
    """
    Wrapper for annoy to use the FLANN api
    """
    def __init__(self):
        self.ann = None
        self.params = {
            'trees': 8,
            'checks': 512,
        }

    def build_index(self, dvecs, **kwargs):
        import annoy
        self.params.update(kwargs)
        self.ann = annoy.AnnoyIndex(f=dvecs.shape[1], metric='euclidean')
        for i, dvec in enumerate(dvecs):
            self.ann.add_item(i, dvec)
        self.ann.build(n_trees=self.params['trees'])

    def nn_index(self, qvecs, num_neighbs, checks=None):
        if checks is None:
            checks = self.params['checks']
        idxs = np.empty((len(qvecs), num_neighbs), dtype=int)
        dists = np.empty((len(qvecs), num_neighbs), dtype=float)
        for i, qvec in enumerate(qvecs):
            idxs[i], dists[i] = self.ann.get_nns_by_vector(
                qvec, n=num_neighbs, search_k=checks, include_distances=True)
        return idxs, dists


def test_annoy():
    from vtool_ibeis import demodata
    import annoy
    import utool
    qvecs = demodata.testdata_dummy_sift(2 * 1000)
    dvecs = demodata.testdata_dummy_sift(100 * 1000)
    dim = dvecs.shape[1]

    checks = 200
    num_neighbs = 10
    num_trees = 8

    trials = 10

    for timer in utool.Timerit(trials, label='build annoy'):
        with timer:
            ann = annoy.AnnoyIndex(dim, metric='euclidean')
            for i, vec in enumerate(dvecs):
                ann.add_item(i, vec)
            ann.build(n_trees=num_trees)

    for timer in utool.Timerit(trials, label='annoy query'):
        with timer:
            for qvec in qvecs:
                ann.get_nns_by_vector(qvec, n=num_neighbs, search_k=checks,
                                      include_distances=True)

    # ---------------

    for timer in utool.Timerit(trials, label='build flann'):
        with timer:
            flann = FLANN_CLS()
            flann.build_index(dvecs, algorithm='kdtree', trees=num_trees,
                              checks=checks, cores=1)

    for timer in utool.Timerit(trials, label='flann query'):
        with timer:
            flann.nn_index(qvecs, num_neighbs, checks=checks)

    # ---------------

    for timer in utool.Timerit(trials, label='build annoy wrapper'):
        with timer:
            index = AnnoyWrapper()
            index.build_index(dvecs, trees=num_trees, checks=checks)

    for timer in utool.Timerit(trials, label='query annoy wrapper'):
        with timer:
            index.nn_index(qvecs, num_neighbs, checks=checks)


def test_cv2_flann():
    """
    Ignore:
        [name for name in dir(cv2) if 'create' in name.lower()]
        [name for name in dir(cv2) if 'stereo' in name.lower()]

        ut.grab_zipped_url('https://priithon.googlecode.com/archive/a6117f5e81ec00abcfb037f0f9da2937bb2ea47f.tar.gz', download_dir='.')
    """
    import cv2
    from vtool_ibeis import demodata
    import plottool_ibeis as pt
    import vtool_ibeis as vt
    # img1 = vt.imread(ut.grab_test_imgpath('easy1.png'))
    # img2 = vt.imread(ut.grab_test_imgpath('easy2.png'))
    img1 = vt.imread(ut.grab_test_imgpath('tsukuba_l'))
    img2 = vt.imread(ut.grab_test_imgpath('tsukuba_r'))

    stereo = cv2.StereoBM_create(numDisparities=16, blockSize=15)
    disparity = stereo.compute(img1, img2)
    pt.imshow(disparity)
    pt.show()

    #cv2.estima

    flow = cv2.createOptFlow_DualTVL1()
    img1, img2 = vt.convert_image_list_colorspace([img1, img2], 'gray', src_colorspace='bgr')
    img2 = vt.resize(img2, img1.shape[0:2][::-1])
    out = img1.copy()
    flow.calc(img1, img2, out)

    orb = cv2.ORB_create()
    kp1, vecs1 = orb.detectAndCompute(img1, None)
    kp2, vecs2 = orb.detectAndCompute(img2, None)

    detector = cv2.FeatureDetector_create("SIFT")
    descriptor = cv2.DescriptorExtractor_create("SIFT")

    skp = detector.detect(img1)
    skp, sd = descriptor.compute(img1, skp)

    tkp = detector.detect(img2)
    tkp, td = descriptor.compute(img2, tkp)

    out = img1.copy()
    cv2.drawKeypoints(img1, kp1, outImage=out)
    pt.imshow(out)

    vecs1 = demodata.testdata_dummy_sift(10)
    vecs2 = demodata.testdata_dummy_sift(10)  # NOQA

    FLANN_INDEX_KDTREE = 0  # bug: flann enums are missing
    #flann_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=4)
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)   # or pass empty dictionary
    flann = cv2.FlannBasedMatcher(index_params, search_params)  # NOQA

    cv2.flann.Index(vecs1, index_params)

    #cv2.FlannBasedMatcher(flann_params)

    cv2.flann.Index(vecs1, flann_params)  # NOQA

    #def match_flann(desc1, desc2, r_threshold = 0.6):
    #    flann = cv2.flann_Index(desc2, flann_params)
    #    idx2, dist = flann.knnSearch(desc1, 2, params = {}) # bug: need to provide empty dict
    #    mask = dist[:,0] / dist[:,1] < r_threshold
    #    idx1 = np.arange(len(desc1))
    #    pairs = np.int32( zip(idx1, idx2[:,0]) )
    #    return pairs[mask]


def ann_flann_once(dpts, qpts, num_neighbors, flann_params={}):
    """
    Finds the approximate nearest neighbors of qpts in dpts

    CommandLine:
        xdoctest -m ~/code/vtool/vtool_ibeis/nearest_neighbors.py ann_flann_once:0

    Example0:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> np.random.seed(1)
        >>> dpts = np.random.randint(0, 255, (5, 128)).astype(np.uint8)
        >>> qpts = np.random.randint(0, 255, (5, 128)).astype(np.uint8)
        >>> qx2_dx, qx2_dist = ann_flann_once(dpts, qpts, 2)
        >>> import ubelt as ub
        >>> result = ub.repr2((qx2_dx.T, qx2_dist.T), precision=2, with_dtype=True, nl=2)
        >>> print(result)
        (
            np.array([[3, 3, 3, 3, 0],
                      [2, 0, 1, 4, 4]], dtype=np.int32),
            np.array([[1037329., 1235876., 1168550., 1286435., 1075507.],
                      [1038324., 1243690., 1304896., 1320598., 1369036.]], dtype=np.float32),
        )


    Example1:
        >>> # ENABLE_DOCTEST
        >>> # Test upper bounds on sift descriptors
        >>> # SeeAlso distance.understanding_pseudomax_props
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> # get points on unit sphere
        >>> nDpts = 5000 # 5
        >>> nQpts = 10000 # 10
        >>> dpts = vt.normalize_rows(np.random.rand(nDpts, 128))
        >>> qpts = vt.normalize_rows(np.random.rand(nQpts, 128))
        >>> qmag = np.sqrt(np.power(qpts, 2).sum(1))
        >>> dmag = np.sqrt(np.power(dpts, 2).sum(1))
        >>> assert np.all(np.allclose(qmag, 1)), 'not on unit sphere'
        >>> assert np.all(np.allclose(dmag, 1)), 'not on unit sphere'
        >>> # cast to uint8
        >>> uint8_max = 512  # hack
        >>> uint8_min = 0  # hack
        >>> K = 100 # 2

        >>> qpts8 = np.clip(np.round(qpts * uint8_max), uint8_min, uint8_max).astype(np.uint8)
        >>> dpts8 = np.clip(np.round(dpts * uint8_max), uint8_min, uint8_max).astype(np.uint8)
        >>> qmag8 = np.sqrt(np.power(qpts8.astype(np.float32), 2).sum(1))
        >>> dmag8 = np.sqrt(np.power(dpts8.astype(np.float32), 2).sum(1))
        >>> # test
        >>> qx2_dx, qx2_dist = ann_flann_once(dpts8, qpts8, K)
        >>> biggest_dist = np.sqrt(qx2_dist.max())
        >>> print('biggest_dist = %r' % (biggest_dist))
        >>> # Get actual distance by hand
        >>> hand_dist = np.sum((qpts8 - dpts8[qx2_dx.T[0]]) ** 2, 0)
        >>> # Seems like flann returns squared distance. makes sense
        >>> result = ub.hash_data(repr((qx2_dx, qx2_dist)))
        >>> print(result)

     Example2:
        >>> # Build theoretically maximally distant vectors
        >>> b = 512
        >>> D = 128
        >>> x = np.sqrt((float(b) ** 2) / float(D - 1))
        >>> dpts = np.ones((2, 128)) * x
        >>> qpts = np.zeros((2, 128))
        >>> dpts[:, 0] = 0
        >>> qpts[:, 0] = 512
        >>> qpts[:, 0::2] = 1
        >>> dpts[:, 1::2] = 1
        >>> qpts[:, 1::2] = 0
        >>> dpts[:, 0::2] = 0
        >>> qmag = np.sqrt(np.power(qpts.astype(np.float64), 2).sum(1))
        >>> dmag = np.sqrt(np.power(dpts.astype(np.float64), 2).sum(1))
        >>> # FIX TO ACTUALLY BE AT THE RIGHT NORM
        >>> dpts = (dpts * (512 / np.linalg.norm(dpts, axis=1))[:, None]).astype(np.float32)
        >>> qpts = (qpts * (512 / np.linalg.norm(qpts, axis=1))[:, None]).astype(np.float32)
        >>> print(np.linalg.norm(dpts))
        >>> print(np.linalg.norm(qpts))
        >>> dist = np.sqrt(np.sum((qpts - dpts) ** 2, 1))
        >>> # Because of norm condition another maximally disant pair of vectors
        >>> # is [1, 0, 0, ... 0] and [0, 1, .. 0, 0, 0]
        >>> # verifythat this gives you same dist.
        >>> dist2 = np.sqrt((512 ** 2 + 512 ** 2))
        >>> print(dist2)
        >>> print(dist)
    """
    # qx2_dx   = query_index -> nearest database index
    # qx2_dist = query_index -> distance
    # import cv2

    flann = FLANN_CLS()
    # obj = cv2.flann_Index()
    if 'algorithm' not in flann_params:
        flann_params['algorithm'] = 0
        # flann_params['trees'] = 5
    flann.build_index(dpts, **flann_params)
    (qx2_dx, qx2_dist) = flann.nn_index(qpts, num_neighbors)
    # flann.build_index(dpts, **flann_params)
    # (qx2_dx, qx2_dist) = .nn_index(qpts, num_neighbors)
    return (qx2_dx, qx2_dist)


def assign_to_centroids(dpts, qpts, num_neighbors=1, flann_params={}):
    """ Helper for akmeans """
    (qx2_dx, qx2_dist) = FLANN_CLS().nn_index(
        dpts, qpts, num_neighbors, **flann_params)
    return qx2_dx


def get_flann_params_cfgstr(flann_params):
    if True:
        # Ensure consistent ordering
        flann_vals = list(flann_params.values())
        flann_keys = list(flann_params.keys())
        # reverse to maintain backwards compatibility
        flann_valsig_ = str(ut.sortedby(flann_vals, flann_keys, reverse=True))
    else:
        flann_valsig_ = str(list(flann_params.values()))
    flann_valsig = ut.remove_chars(flann_valsig_, ', \'[]')
    return flann_valsig


def get_flann_cfgstr(dpts, flann_params, cfgstr='', use_params_hash=True,
                     use_data_hash=True):
    """

    CommandLine:
        python -m vtool_ibeis.nearest_neighbors --test-get_flann_cfgstr

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> rng = np.random.RandomState(1)
        >>> dpts = rng.randint(0, 255, (10, 128)).astype(np.uint8)
        >>> cache_dir = '.'
        >>> cfgstr = '_FEAT(alg=heshes)'
        >>> flann_params = get_kdtree_flann_params()
        >>> result = get_flann_cfgstr(dpts, flann_params, cfgstr)
        >>> print(result)
        _FEAT(alg=heshes)_FLANN(4kdtree)_DPTS((10,128)xxaotseonmfjkzcr)
    """
    flann_cfgstr = cfgstr
    if use_params_hash:
        flann_valsig = get_flann_params_cfgstr(flann_params)
        flann_cfgstr += '_FLANN(' + flann_valsig + ')'
    # Generate a unique filename for dpts and flann parameters
    if use_data_hash:
        # flann is dependent on the dpts
        data_hashstr = ut.hashstr_arr27(dpts, '_DPTS')
        flann_cfgstr += data_hashstr
    return flann_cfgstr


def get_flann_fpath(dpts, cache_dir='default', cfgstr='', flann_params={},
                    use_params_hash=True, use_data_hash=True, appname='vtool_ibeis',
                    verbose=True):
    """ returns filepath for flann index """
    if cache_dir == 'default':
        if verbose:
            print('[flann] using default cache dir')
        cache_dir = ub.ensure_app_cache_dir(appname)
        ub.ensuredir(cache_dir)
    flann_cfgstr = get_flann_cfgstr(dpts, flann_params, cfgstr,
                                    use_params_hash=use_params_hash,
                                    use_data_hash=use_data_hash)
    if verbose:
        print('...flann_cache cfgstr = %r: ' % flann_cfgstr)
    # Append any user labels
    flann_fname = 'flann_index' + flann_cfgstr + '.flann'
    flann_fpath = normpath(join(cache_dir, flann_fname))
    return flann_fpath


def flann_cache(dpts, cache_dir='default', cfgstr='', flann_params={},
                use_cache=True, save=True, use_params_hash=True,
                use_data_hash=True, appname='vtool_ibeis', verbose=None):
    """
    Tries to load a cached flann index before doing anything
    from vtool_ibeis.nn
    """
    if verbose is None:
        verbose = int(ut.NOT_QUIET)
    if verbose is True:
        verbose = 2
    if verbose > 1:
        print('+--- START CACHED FLANN INDEX ')
    if len(dpts) == 0:
        raise ValueError(
            'cannot build flann when len(dpts) == 0. (prevents a segfault)')
    flann_fpath = get_flann_fpath(dpts, cache_dir, cfgstr, flann_params,
                                  use_params_hash=use_params_hash,
                                  use_data_hash=use_data_hash, appname=appname,
                                  verbose=verbose)
    # Load the index if it exists
    flann = FLANN_CLS()
    flann.flann_fpath = flann_fpath

    if use_cache and exists(flann_fpath):
        try:
            flann.load_index(flann_fpath, dpts)
            if verbose > 0:
                print('...flann cache hit: %d vectors' % (len(dpts)))
            if verbose > 1:
                print('L___ END FLANN INDEX ')
            return flann
        except Exception as ex:
            ut.printex(ex, '... cannot load index', iswarning=True)
    # Rebuild the index otherwise
    if verbose > 0:
        print('...flann cache miss.')
    num_dpts = len(dpts)
    if flann is None:
        flann = FLANN_CLS()
    if verbose > 1 or (verbose > 0 and num_dpts > 1E6):
        print('...building kdtree over %d points (this may take a sec).' % num_dpts)
    if num_dpts == 0:
        print('WARNING: CANNOT BUILD FLANN INDEX OVER 0 POINTS. THIS MAY BE A SIGN OF A DEEPER ISSUE')
        return flann

    flann.build_index(dpts, **flann_params)
    if verbose > 1:
        print('flann.save_index(%r)' % ut.path_ndir_split(flann_fpath, n=2))
    if save:
        flann.save_index(flann_fpath)
    if verbose > 1:
        print('L___ END CACHED FLANN INDEX ')
    return flann


def flann_augment(dpts, new_dpts, cache_dir, cfgstr, new_cfgstr, flann_params,
                  use_cache=True, save=True):
    """
    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata  # NOQA
        >>> dpts = demodata.get_dummy_dpts(ut.get_nth_prime(10))
        >>> new_dpts = demodata.get_dummy_dpts(ut.get_nth_prime(9))
        >>> cache_dir = ut.get_app_resource_dir('vtool_ibeis')
        >>> cfgstr = '_testcfg'
        >>> new_cfgstr = '_new_testcfg'
        >>> flann_params = get_kdtree_flann_params()
        >>> use_cache = False
        >>> save = False
    """
    flann = flann_cache(dpts, cache_dir, cfgstr, flann_params)
    flann.add_points(new_dpts)
    if save:
        aug_dpts = np.vstack((dpts, new_dpts))
        new_flann_fpath = get_flann_fpath(
            aug_dpts, cache_dir, new_cfgstr, flann_params)
        flann.save_index(new_flann_fpath)
    return flann


def get_kdtree_flann_params():
    flann_params = {
        'algorithm': 'kdtree',
        'trees': 4
    }
    return flann_params


def get_flann_params(algorithm='kdtree', **kwargs):
    """
    Returns flann params that are relvant tothe algorithm

    References:
        http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_manual-1.8.4.pdf

    Args:
        algorithm (str): (default = 'kdtree')

    Returns:
        dict: flann_params

    CommandLine:
        python -m vtool_ibeis.nearest_neighbors --test-get_flann_params --algo=kdtree
        python -m vtool_ibeis.nearest_neighbors --test-get_flann_params --algo=kmeans

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> algorithm = ut.get_argval('--algo', default='kdtree')
        >>> flann_params = get_flann_params(algorithm)
        >>> result = ('flann_params = %s' % (ub.repr2(flann_params),))
        >>> print(result)
    """
    _algorithm_options = [
        'linear',
        'kdtree',
        'kmeans',
        'composite',
        'kdtree_single'
    ]
    _centersinit_options = [
        'random',
        'gonzales',
        'kmeanspp',
    ]
    # Search params (for all algos)
    assert algorithm in _algorithm_options
    flann_params = {
        'algorithm': algorithm
    }
    if algorithm != 'linear':
        flann_params.update({
            'random_seed': -1
        })
    if algorithm in ['kdtree', 'composite']:
        # kdtree index parameters
        flann_params.update({
            'algorithm': _algorithm_options[1],
            'trees': 4,
            'checks': 32,  # how many leafs (features) to check in one search
        })
    elif algorithm in ['kmeans', 'composite']:
        # Kmeans index parametrs
        flann_params.update({
            'branching': 32,
            'iterations': 5,
            'centers_init': _centersinit_options[2],
            'cb_index': 0.5,  # cluster boundary index for searching kmeanms tree
            'checks': 32,  # how many leafs (features) to check in one search
        })
    elif algorithm == 'autotuned':
        flann_params.update({
            'algorithm'        : 'autotuned',
            'target_precision' : .01,    # precision desired (used for autotuning, -1 otherwise)
            'build_weight'     : 0.01,   # build tree time weighting factor
            'memory_weight'    : 0.0,    # index memory weigthing factor
            'sample_fraction'  : 0.001,  # what fraction of the dataset to use for autotuning
        })
    elif algorithm == 'lsh':
        flann_params.update({
            'table_number_': 12,
            'key_size_': 20,
            'multi_probe_level_': 2,
        })

    flann_params = ut.update_existing(flann_params, kwargs, assert_exists=True)
    return flann_params


def tune_flann(dpts,
               target_precision=.90,
               build_weight=0.50,
               memory_weight=0.00,
               sample_fraction=0.01):
    r"""

    References:
        http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_pami2014.pdf
        http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_manual-1.8.4.pdf
        http://docs.opencv.org/trunk/modules/flann/doc/flann_fast_approximate_nearest_neighbor_search.html

    Math::
        cost of an algorithm is:

        LaTeX:
            \cost = \frac
                {\search + build_weight * \build }
                { \minoverparams( \search + build_weight \build)} +
                memory_weight * \memory

    Args:
        dpts (ndarray):

        target_precision (float): number between 0 and 1 representing desired
            accuracy. Higher values are more accurate.

        build_weight (float): importance weight given to minimizing build time
            relative to search time. This number can range from 0 to infinity.
            typically because building is a more complex computation you want
            to keep the number relatively low, (less than 1) otherwise you'll
            end up getting a linear search (no build time).

        memory_weight (float): Importance of memory relative to total speed.
            A value less than 1 gives more importance to the time spent and a
            value greater than 1 gives more importance to the memory usage.

        sample_fraction (float): number between 0 and 1 representing the
            fraction of the input data to use in the optimization. A higher
            number uses more data.

    Returns:
        dict: tuned_params

    CommandLine:
        python -m vtool_ibeis.nearest_neighbors --test-tune_flann

    """
    with ut.Timer('tuning flann'):
        print('Autotuning flann with %d %dD vectors' % (dpts.shape[0], dpts.shape[1]))
        print('a sample of %d vectors will be used' % (int(dpts.shape[0] * sample_fraction)))
        flann = FLANN_CLS()
        #num_data = len(dpts)
        flann_atkwargs = dict(algorithm='autotuned',
                              target_precision=target_precision,
                              build_weight=build_weight,
                              memory_weight=memory_weight,
                              sample_fraction=sample_fraction)
        suffix = repr(flann_atkwargs)
        badchar_list = ',{}\': '
        for badchar in badchar_list:
            suffix = suffix.replace(badchar, '')
        print('flann_atkwargs:')
        print(ub.repr2(flann_atkwargs))
        print('starting optimization')
        tuned_params = flann.build_index(dpts, **flann_atkwargs)
        print('finished optimization')

        # The algorithm is sometimes returned as default which is
        # very unuseful as the default name is embeded in the pyflann
        # module where most would not care to look. This finds the default
        # name for you.
        for key in ['algorithm', 'centers_init', 'log_level']:
            val = tuned_params.get(key, None)
            if val == 'default':
                dict_ = pyflann.FLANNParameters._translation_[key]
                other_algs = ut.dict_find_other_sameval_keys(dict_, 'default')
                assert len(other_algs) == 1, 'more than 1 default for key=%r' % (key,)
                tuned_params[key] = other_algs[0]

        common_params = [
            'algorithm',
            'checks',
        ]
        relevant_params_dict = dict(
            linear=['algorithm'],
            #---
            kdtree=[
                'trees'
            ],
            #---
            kmeans=[
                'branching',
                'iterations',
                'centers_init',
                'cb_index',
            ],
            #---
            lsh=[
                'table_number',
                'key_size',
                'multi_probe_level',
            ],
        )
        relevant_params_dict['composite'] = relevant_params_dict['kmeans'] + relevant_params_dict['kdtree'] + common_params
        relevant_params_dict['kmeans'] += common_params
        relevant_params_dict['kdtree'] += common_params
        relevant_params_dict['lsh'] += common_params

        #kdtree_single_params = [
        #    'leaf_max_size',
        #]
        #other_params = [
        #    'build_weight',
        #    'sorted',
        #]
        out_file = 'flann_tuned' + suffix
        ut.write_to(out_file, ub.repr2(tuned_params, sorted_=True, newlines=True))
        flann.delete_index()
        if tuned_params['algorithm'] in relevant_params_dict:
            print('relevant_params=')
            relevant_params = relevant_params_dict[tuned_params['algorithm']]
            print(ub.repr2(ut.dict_subset(tuned_params, relevant_params),
                              sorted_=True, newlines=True))
            print('irrelevant_params=')
            print(ub.repr2(ut.dict_setdiff(tuned_params, relevant_params),
                              sorted_=True, newlines=True))
        else:
            print('unknown tuned algorithm=%r' % (tuned_params['algorithm'],))

        print('all_tuned_params=')
        print(ub.repr2(tuned_params, sorted_=True, newlines=True))
    return tuned_params


def flann_index_time_experiment():
    r"""

    Shows a plot of how long it takes to build a flann index for a given number of KD-trees

    CommandLine:
        python -m vtool_ibeis.nearest_neighbors --test-flann_index_time_experiment

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> result = flann_index_time_experiment()
        >>> print(result)
    """
    import vtool_ibeis as vt
    import itertools

    class TestDataPool(object):
        """
        Perform only a few allocations of test data
        """
        def __init__(self):
            self.num = 10000
            self.data_pool = None
            self.alloc_pool(1000000)

        def alloc_pool(self, num):
            print('[alloc] num = %r' % (num,))
            self.num = num
            self.data_pool = vt.demodata.testdata_dummy_sift(num)
            print('[alloc] object size ' + ut.get_object_size_str(self.data_pool, 'data_pool'))

        def get_testdata(self, num):
            if len(self.data_pool) < num:
                self.alloc_pool(2 * self.num)
            return self.data_pool[0:num]

    pool = TestDataPool()

    def get_buildtime_data(**kwargs):
        flann_params = vt.get_flann_params(**kwargs)
        print('flann_params = %r' % (ub.repr2(flann_params),))
        data_list = []
        num = 1000
        print('-----')
        for count in ut.ProgIter(itertools.count(), length=-1, freq=1,
                                 adjust=False):
            num = int(num * 1.2)
            print('num = %r' % (num,))
            #if num > 1E6:
            #    break
            data = pool.get_testdata(num)
            print('object size ' + ut.get_object_size_str(data, 'data'))
            flann = FLANN_CLS(**flann_params)
            with ut.Timer(verbose=False) as t:
                flann.build_index(data)
            print('t.ellapsed = %r' % (t.ellapsed,))
            if t.ellapsed > 5 or count > 1000:
                break
            data_list.append((count, num, t.ellapsed))
            print('-----')
        return data_list, flann_params

    data_list1, params1 = get_buildtime_data(trees=1)

    data_list2, params2 = get_buildtime_data(trees=2)

    data_list4, params4 = get_buildtime_data(trees=4)

    data_list8, params8 = get_buildtime_data(trees=8)

    data_list16, params16 = get_buildtime_data(trees=16)

    import plottool_ibeis as pt

    def plotdata(data_list):
        count_arr = ut.get_list_column(data_list, 1)
        time_arr  = ut.get_list_column(data_list, 2)
        pt.plot2(count_arr, time_arr, marker='-o', equal_aspect=False,
                 x_label='num_vectors', y_label='FLANN build time')

    plotdata(data_list1)
    plotdata(data_list2)
    plotdata(data_list4)
    plotdata(data_list8)
    plotdata(data_list16)

    pt.iup()


def invertible_stack(vecs_list, label_list):
    """
    Stacks descriptors into a flat structure and returns inverse mapping from
    flat database descriptor indexes (dx) to annotation ids (label) and feature
    indexes (fx). Feature indexes are w.r.t. annotation indexes.

    Output:
        idx2_desc - flat descriptor stack
        idx2_label  - inverted index into annotations
        idx2_fx   - inverted index into features

    # Example with 2D Descriptors

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyflann_ibeis)
        >>> from vtool_ibeis.nearest_neighbors import *  # NOQA
        >>> DESC_TYPE = np.uint8
        >>> label_list  = [1, 2, 3, 4, 5]
        >>> vecs_list = [
        ...     np.array([[0, 0], [0, 1]], dtype=DESC_TYPE),
        ...     np.array([[5, 3], [2, 30], [1, 1]], dtype=DESC_TYPE),
        ...     np.empty((0, 2), dtype=DESC_TYPE),
        ...     np.array([[5, 3], [2, 30], [1, 1]], dtype=DESC_TYPE),
        ...     np.array([[3, 3], [42, 42], [2, 6]], dtype=DESC_TYPE),
        ...     ]
        >>> idx2_vec, idx2_label, idx2_fx = invertible_stack(vecs_list, label_list)
        >>> print(repr(idx2_vec.T))
        array([[ 0,  0,  5,  2,  1,  5,  2,  1,  3, 42,  2],
               [ 0,  1,  3, 30,  1,  3, 30,  1,  3, 42,  6]], dtype=uint8)
        >>> print(repr(idx2_label))
        array([1, 1, 2, 2, 2, 4, 4, 4, 5, 5, 5])
        >>> print(repr(idx2_fx))
        array([0, 1, 0, 1, 2, 0, 1, 2, 0, 1, 2])
    """
    # INFER DTYPE? dtype = vecs_list[0].dtype
    # Build inverted index of (label, fx) pairs
    nFeats = sum(list(map(len, vecs_list)))
    nFeat_iter = map(len, vecs_list)
    label_nFeat_iter = zip(label_list, map(len, vecs_list))
    # generate featx inverted index for each feature in each annotation
    _ax2_fx = [list(range(nFeat)) for nFeat in nFeat_iter]
    # generate label inverted index for each feature in each annotation
    _ax2_label = [[label] * nFeat for (label, nFeat) in label_nFeat_iter]
    # Flatten generators into the inverted index
    _flatlabels = ub.flatten(_ax2_label)
    _flatfeatxs = ub.flatten(_ax2_fx)

    idx2_label = np.fromiter(_flatlabels, np.int32, nFeats)
    idx2_fx = np.fromiter(_flatfeatxs, np.int32, nFeats)
    # Stack vecsriptors into numpy array corresponding to inverted inexed
    # This might throw a MemoryError
    idx2_vec = np.vstack(vecs_list)
    return idx2_vec, idx2_label, idx2_fx


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.nearest_neighbors
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import utool as ut
import ubelt as ub
import six


def extract_feature_from_patch(patch):
    import pyhesaff
    import numpy as np
    import vtool_ibeis as vt
    patch = vt.rectify_to_uint8(patch)
    patch = vt.rectify_to_square(patch)
    patch_list = np.ascontiguousarray(patch[None, :])
    vec = pyhesaff.extract_desc_from_patches(patch_list)[0]
    return vec


def extract_features(img_or_fpath, feat_type='hesaff+sift', **kwargs):
    r"""
    calls pyhesaff's main driver function for detecting hessian affine keypoints.
    extra parameters can be passed to the hessian affine detector by using
    kwargs.

    Args:
        img_or_fpath (str): image file path on disk
        use_adaptive_scale (bool):
        nogravity_hack (bool):

    Returns:
        tuple : (kpts, vecs)

    CommandLine:
        python -m vtool_ibeis.features --test-extract_features
        python -m vtool_ibeis.features --test-extract_features --show
        python -m vtool_ibeis.features --test-extract_features --feat-type=hesaff+siam128 --show
        python -m vtool_ibeis.features --test-extract_features --feat-type=hesaff+siam128 --show
        python -m vtool_ibeis.features --test-extract_features --feat-type=hesaff+siam128 --show --no-affine-invariance

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.features import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> img_fpath = ut.grab_test_imgpath(ut.get_argval('--fname', default='astro'))
        >>> imgBGR = vt.imread(img_fpath)
        >>> feat_type = ub.argval('--feat_type', default='hesaff+sift')
        >>> import pyhesaff
        >>> kwargs = ut.parse_dict_from_argv(pyhesaff.get_hesaff_default_params())
        >>> # execute function
        >>> #(kpts, vecs) = extract_features(img_fpath)
        >>> (kpts, vecs) = extract_features(imgBGR, feat_type, **kwargs)
        >>> # verify results
        >>> result = str((kpts, vecs))
        >>> print(result)
        >>> # Show keypoints
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> #pt.figure(fnum=1, doclf=True, docla=True)
        >>> #pt.imshow(imgBGR)
        >>> #pt.draw_kpts2(kpts, ori=True)
        >>> pt.interact_keypoints.ishow_keypoints(imgBGR, kpts, vecs, ori=True, ell_alpha=.4, color='distinct')
        >>> pt.show_if_requested()
    """
    import pyhesaff
    if feat_type == 'hesaff+sift':
        #(kpts, vecs) = pyhesaff.detect_feats(img_fpath, **kwargs)
        (kpts, vecs) = pyhesaff.detect_feats2(img_or_fpath, **kwargs)
    elif feat_type == 'hesaff+siam128':
        # hacky
        from ibeis_cnn import _plugin
        (kpts, sift) = pyhesaff.detect_feats2(img_or_fpath, **kwargs)
        if isinstance(img_or_fpath, six.string_types):
            import vtool_ibeis as vt
            img_or_fpath = vt.imread(img_or_fpath)
        vecs_list = _plugin.extract_siam128_vecs([img_or_fpath], [kpts])
        vecs = vecs_list[0]
        pass
    else:
        raise AssertionError('Unknown feat_type=%r' % (feat_type,))
    return (kpts, vecs)


def get_extract_features_default_params():
    r"""
    Returns:
        dict:

    CommandLine:
        python -m vtool_ibeis.features --test-get_extract_features_default_params

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.features import *  # NOQA
        >>> # build test data
        >>> # execute function
        >>> param_dict = get_extract_features_default_params()
        >>> result = ub.repr2(param_dict)
        >>> # verify results
        >>> print(result)
    """
    import pyhesaff
    param_dict = pyhesaff.get_hesaff_default_params()
    return param_dict


def detect_opencv_keypoints():
    import cv2
    import vtool_ibeis as vt
    import numpy as np  # NOQA

    img_fpath = ut.grab_test_imgpath(ub.argval('--fname', default='astro'))
    imgBGR = vt.imread(img_fpath)
    imgGray = cv2.cvtColor(imgBGR, cv2.COLOR_BGR2GRAY)

    def from_cv2_kpts(cv2_kp):
        kp = (cv2_kp.pt[0], cv2_kp.pt[1], cv2_kp.size, 0, cv2_kp.size, cv2_kp.angle)
        return kp

    print('\n'.join(ut.search_module(cv2, 'create', recursive=True)))

    detect_factory = {
        #'BLOB': cv2.SimpleBlobDetector_create,
        #'HARRIS' : HarrisWrapper.create,
        #'SIFT': cv2.xfeatures2d.SIFT_create,  # really DoG
        'SURF': cv2.xfeatures2d.SURF_create,  # really harris corners
        'MSER': cv2.MSER_create,
        #'StarDetector_create',

    }

    extract_factory = {
        'SIFT': cv2.xfeatures2d.SIFT_create,
        'SURF': cv2.xfeatures2d.SURF_create,
        #'DAISY': cv2.xfeatures2d.DAISY_create,
        'FREAK': cv2.xfeatures2d.FREAK_create,
        #'LATCH': cv2.xfeatures2d.LATCH_create,
        #'LUCID': cv2.xfeatures2d.LUCID_create,
        #'ORB': cv2.ORB_create,
    }
    mask = None

    type_to_kpts = {}
    type_to_desc = {}

    key = 'BLOB'
    key = 'MSER'

    for key in detect_factory.keys():
        factory = detect_factory[key]
        extractor = factory()

        # For MSERS need to adapt shape and then convert into a keypoint repr
        if hasattr(extractor, 'detectRegions'):
            # bboxes are x,y,w,h
            regions, bboxes = extractor.detectRegions(imgGray)
            # ellipse definition from [Fitzgibbon95]
            # http://www.bmva.org/bmvc/1995/bmvc-95-050.pdf p518
            # ell = [c_x, c_y, R_x, R_y, theta]
            # (cx, cy) = conic center
            # Rx and Ry = conic radii
            # theta is the counterclockwise angle
            fitz_ellipses = [cv2.fitEllipse(mser) for mser in regions]

            # http://answers.opencv.org/question/19015/how-to-use-mser-in-python/
            #hulls = [cv2.convexHull(p.reshape(-1, 1, 2)) for p in regions]
            #hull_ells = [cv2.fitEllipse(hull[:, 0]) for hull in hulls]
            kpts_ = []
            for ell in fitz_ellipses:
                ((cx, cy), (rx, ry), degrees) = ell
                theta = np.radians(degrees)  # opencv lives in radians
                S = vt.scale_mat3x3(rx, ry)
                T = vt.translation_mat3x3(cx, cy)
                R = vt.rotation_mat3x3(theta)
                #R = np.eye(3)
                invVR = T.dot(R.dot(S))
                kpt = vt.flatten_invV_mats_to_kpts(np.array([invVR]))[0]
                kpts_.append(kpt)
            kpts_ = np.array(kpts_)

        tt = ut.tic('Computing %r keypoints' % (key,))
        try:
            cv2_kpts = extractor.detect(imgGray, mask)
        except Exception as ex:
            ut.printex(ex, 'Failed to computed %r keypoints' % (key,), iswarning=True)
            pass
        else:
            ut.toc(tt)
            type_to_kpts[key] = cv2_kpts

    print(list(type_to_kpts.keys()))
    print(ut.depth_profile(list(type_to_kpts.values())))
    print('type_to_kpts = ' + ub.repr2(type_to_kpts, truncate=True))

    cv2_kpts = type_to_kpts['MSER']
    kp = cv2_kpts[0]  # NOQA
    #cv2.fitEllipse(cv2_kpts[0])
    cv2_kpts = type_to_kpts['SURF']

    for key in extract_factory.keys():
        factory = extract_factory[key]
        extractor = factory()
        tt = ut.tic('Computing %r descriptors' % (key,))
        try:
            filtered_cv2_kpts, desc = extractor.compute(imgGray, cv2_kpts)
        except Exception as ex:
            ut.printex(ex, 'Failed to computed %r descriptors' % (key,), iswarning=True)
            pass
        else:
            ut.toc(tt)
            type_to_desc[key] = desc

    print(list(type_to_desc.keys()))
    print(ut.depth_profile(list(type_to_desc.values())))
    print('type_to_desc = ' + ub.repr2(type_to_desc, truncate=True))


def test_mser():
    import cv2
    import vtool_ibeis as vt
    import plottool_ibeis as pt
    import numpy as np
    pt.qt4ensure()
    class Keypoints(ut.NiceRepr):
        """
        Convinence class for dealing with keypoints
        """
        def __init__(self, kparr, info=None):
            self.kparr = kparr
            if info is None:
                info = {}
            self.info = info

        def add_info(self, key, val):
            self.info[key] = val

        def __nice__(self):
            return ' ' + str(len(self.kparr))

        @property
        def scale(self):
            return vt.get_scales(self.kparr)

        @property
        def eccentricity(self):
            return vt.get_kpts_eccentricity(self.kparr)

        def compress(self, flags, inplace=False):
            subarr = self.kparr.compress(flags, axis=0)
            info = {key: list(ub.compress(val, flags)) for key, val in self.info.items()}
            return Keypoints(subarr, info)

    img_fpath = ut.grab_test_imgpath(ub.argval('--fname', default='astro'))
    imgBGR = vt.imread(img_fpath)
    imgGray = cv2.cvtColor(imgBGR, cv2.COLOR_BGR2GRAY)
    # http://docs.opencv.org/master/d3/d28/classcv_1_1MSER.html#gsc.tab=0
    # http://stackoverflow.com/questions/17647500/exact-meaning-of-the-parameters-given-to-initialize-mser-in-opencv-2-4-x
    factory = cv2.MSER_create
    img_area = np.product(np.array(vt.get_size(imgGray)))
    _max_area = (img_area // 10)
    _delta = 8
    _min_diversity = .5

    extractor = factory(_delta=_delta, _max_area=_max_area, _min_diversity=_min_diversity)
    # bboxes are x,y,w,h
    regions, bboxes = extractor.detectRegions(imgGray)
    # ellipse definition from [Fitzgibbon95]
    # http://www.bmva.org/bmvc/1995/bmvc-95-050.pdf p518
    # ell = [c_x, c_y, R_x, R_y, theta]
    # (cx, cy) = conic center
    # Rx and Ry = conic radii
    # theta is the counterclockwise angle
    fitz_ellipses = [cv2.fitEllipse(mser) for mser in regions]

    # http://answers.opencv.org/question/19015/how-to-use-mser-in-python/
    #hulls = [cv2.convexHull(p.reshape(-1, 1, 2)) for p in regions]
    #hull_ells = [cv2.fitEllipse(hull[:, 0]) for hull in hulls]
    invVR_mats = []
    for ell in fitz_ellipses:
        ((cx, cy), (dx, dy), degrees) = ell
        theta = np.radians(degrees)  # opencv lives in radians
        # Convert diameter to radians
        rx = dx / 2
        ry = dy / 2
        S = vt.scale_mat3x3(rx, ry)
        T = vt.translation_mat3x3(cx, cy)
        R = vt.rotation_mat3x3(theta)
        invVR = T.dot(R.dot(S))
        invVR_mats.append(invVR)
    invVR_mats = np.array(invVR_mats)
    #_oris = vt.get_invVR_mats_oris(invVR_mats)
    kpts2_ = vt.flatten_invV_mats_to_kpts(invVR_mats)

    self = Keypoints(kpts2_)
    self.add_info('regions', regions)
    flags = (self.eccentricity < .9)
    #flags = self.scale < np.mean(self.scale)
    #flags = self.scale < np.median(self.scale)
    self = self.compress(flags)
    import plottool_ibeis as pt
    #pt.interact_keypoints.ishow_keypoints(imgBGR, self.kparr, None, ell_alpha=.4, color='distinct', fnum=2)
    #import plottool_ibeis as pt
    vis = imgBGR.copy()

    for region in self.info['regions']:
        vis[region.T[1], region.T[0], :] = 0

    #regions, bbox = mser.detectRegions(gray)
    #hulls = [cv2.convexHull(p.reshape(-1, 1, 2)) for p in self.info['regions']]
    #cv2.polylines(vis, hulls, 1, (0, 255, 0))
    #for region in self.info['regions']:
    #    ell = cv2.fitEllipse(region)
    #    cv2.ellipse(vis, ell, (255))
    pt.interact_keypoints.ishow_keypoints(vis, self.kparr, None, ell_alpha=.4, color='distinct', fnum=2)
    #pt.imshow(vis, fnum=2)
    pt.update()

    #extractor = extract_factory['DAISY']()

    #desc_type_to_dtype = {
    #    cv2.CV_8U: np.uint8,
    #    cv2.CV_8s: np.uint,
    #}
    #def alloc_desc(extractor):
    #    desc_type = extractor.descriptorType()
    #    desc_size = extractor.descriptorSize()
    #    dtype = desc_type_to_dtype[desc_type]
    #    shape = (len(cv2_kpts), desc_size)
    #    desc = np.empty(shape, dtype=dtype)
    #    return desc

    #ut.search_module(cv2, 'array', recursive=True)
    #ut.search_module(cv2, 'freak', recursive=True)
    #ut.search_module(cv2, 'out', recursive=True)

    #cv2_kpts = cv2_kpts[0:2]

    #for key, factory in just_desc_factory_.items():
    #    extractor = factory()
    #    desc = alloc_desc(extractor)
    #    desc = extractor.compute(imgGray, cv2_kpts)
    #    feats[key] = (desc,)
    #    #extractor.compute(imgGray, cv2_kpts, desc)
    #    pass
    #kpts = np.array(list(map(from_cv2_kpts, cv2_kpts)))

    #orb = cv2.ORB()
    #kp1, des1 = orb.detectAndCompute(imgGray, None)
    #blober = cv2.SimpleBlobDetector_create()
    #haris_kpts = cv2.cornerHarris(imgGray, 2, 3, 0.04)

    #[name for name in dir(cv2) if 'mat' in name.lower()]
    #[name for name in dir(cv2.xfeatures2d) if 'desc' in name.lower()]

    #[name for name in dir(cv2) if 'detect' in name.lower()]
    #[name for name in dir(cv2) if 'extract' in name.lower()]
    #[name for name in dir(cv2) if 'ellip' in name.lower()]

    #sift = cv2.xfeatures2d.SIFT_create()
    #cv2_kpts = sift.detect(imgGray)
    #desc = sift.compute(imgGray, cv2_kpts)[1]

    #freak = cv2.xfeatures2d.FREAK_create()
    #cv2_kpts = freak.detect(imgGray)
    #desc = freak.compute(imgGray, cv2_kpts)[1]
    pass


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.features
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import utool as ut
import vtool_ibeis.spatial_verification as sver
import numpy as np
import vtool_ibeis.demodata as demodata
import vtool_ibeis.keypoint as ktool  # NOQA
import vtool_ibeis.linalg as ltool  # NOQA


xy_thresh = ktool.KPTS_DTYPE(.009)
scale_thresh_sqrd = ktool.KPTS_DTYPE(2)
TAU = np.pi * 2.0  # References: tauday.com
ori_thresh = ktool.KPTS_DTYPE(TAU / 4.0)


# def test_sver():
#     nShow = ut.get_argval('--nShow', int, 1)
#     chip1, chip2, kpts1, kpts2, fm = get_dummy_test_vars()
#     demo_sver(chip1, chip2, kpts1, kpts2, fm, nShow)


def demo_sver(chip1, chip2, kpts1, kpts2, fm, nShow=6):
    r"""
    Args:
        chip1 (ndarray[uint8_t, ndim=2]):  annotation image data
        chip2 (ndarray[uint8_t, ndim=2]):  annotation image data
        kpts1 (ndarray[float32_t, ndim=2]):  keypoints
        kpts2 (ndarray[float32_t, ndim=2]):  keypoints
        fm (list):  list of feature matches as tuples (qfx, dfx)
        nShow (int):

    Example0:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:plottool)
        >>> import plottool_ibeis as pt
        >>> # build test data
        >>> nShow = ut.get_argval('--nShow', int, 1)
        >>> chip1, chip2, kpts1, kpts2, fm = get_dummy_test_vars()
        >>> # execute function
        >>> result = demo_sver(chip1, chip2, kpts1, kpts2, fm, nShow)
        >>> # verify results
        >>> print(result)
        >>> pt.show_if_requested()

    Example1:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:plottool)
        >>> import plottool_ibeis as pt
        >>> # build test data
        >>> nShow = ut.get_argval('--nShow', int, 1)
        >>> chip1, chip2, kpts1, kpts2, fm = get_dummy_test_vars1()
        >>> # execute function
        >>> result = demo_sver(chip1, chip2, kpts1, kpts2, fm, nShow)
        >>> # verify results
        >>> print(result)
        >>> pt.show_if_requested()
    """

    xy_thresh_sqrd = ktool.get_diag_extent_sqrd(kpts2) * xy_thresh

    def pack_errors(xy_err, scale_err, ori_err):
        """ makes human readable errors """
        def _pack(bits, errs, thresh):
            return ut.indentjoin(['%5s %f < %f' % (bit, err, thresh) for
                                  (bit, err) in zip(bits, errs)])
        xy_flag = xy_err < xy_thresh_sqrd
        scale_flag = scale_err < scale_thresh_sqrd
        ori_flag = ori_err < ori_thresh
        errors_dict = {
            'xy_err':     _pack(xy_flag, np.sqrt(xy_err), np.sqrt(xy_thresh_sqrd)),
            'scale_err':  _pack(scale_flag, np.sqrt(scale_err), np.sqrt(scale_thresh_sqrd)),
            'ori_err':    _pack(ori_flag, ori_err, ori_thresh),
        }
        return errors_dict

    # Test each affine hypothesis
    #assert kpts1.dtype == ktool.KPTS_DTYPE, 'bad cast somewhere kpts1.dtype=%r' % (kpts1.dtype)
    #assert kpts2.dtype == ktool.KPTS_DTYPE, 'bad cast somewhere kpts2.dtype=%r' % (kpts2.dtype)
    #assert xy_thresh_sqrd.dtype == ktool.KPTS_DTYPE, 'bad cast somewhere
    #xy_thresh_sqrd.dtype=%r' % (xy_thresh_sqrd.dtype)
    aff_hypo_tups = sver.get_affine_inliers(kpts1, kpts2, fm, xy_thresh_sqrd,
                                            scale_thresh_sqrd, ori_thresh)
    inliers_list, errors_list, Aff_mats = aff_hypo_tups

    # Determine best hypothesis
    nInliers_list = np.array(list(map(len, inliers_list)))
    best_mxs = nInliers_list.argsort()[::-1]

    for fnum, mx in enumerate(best_mxs[0:min(len(best_mxs), nShow)]):
        Aff = Aff_mats[mx]
        aff_inliers = inliers_list[mx]
        if ut.get_argflag('--print-error'):
            errors = pack_errors(*errors_list[mx])  # NOQA
            print(ut.repr2(errors, strvals=True))

        homog_inliers, homog_errors, H = sver.get_homography_inliers(kpts1,
                                                                     kpts2, fm,
                                                                     aff_inliers,
                                                                     xy_thresh_sqrd)

        kpts1_At = ktool.transform_kpts(kpts1, Aff)
        kpts1_Ht = ktool.transform_kpts(kpts1, H)
        kpts = kpts1
        M = H

        homog_tup = (homog_inliers, H)
        aff_tup = (aff_inliers, Aff)

        _args = (chip1, chip2, kpts1, kpts2, fm)
        _kw = dict(show_assign=True, show_kpts=True, mx=mx, fnum=fnum * 3)
        from plottool_ibeis import draw_func2 as df2
        from plottool_ibeis import draw_sv
        draw_sv.show_sv(*_args, aff_tup=aff_tup, homog_tup=homog_tup, **_kw)
        #draw_sv.show_sv(*_args, aff_tup=aff_tup, mx=mx, fnum=fnum * 3)
        #draw_sv.show_sv(*_args, homog_tup=homog_tup, mx=mx, fnum=3)

        df2.set_figtitle('# %r inliers (in rects, hypo in bold)' % (nInliers_list[mx],))
    return locals()


def get_dummy_test_vars():
    kpts1 = demodata.perterbed_grid_kpts(seed=12, damping=1.2, dtype=np.float64)
    kpts2 = demodata.perterbed_grid_kpts(seed=24, damping=1.6, dtype=np.float64)
    assert kpts1.dtype == np.float64
    assert kpts2.dtype == np.float64
    chip1 = demodata.get_kpts_dummy_img(kpts1)
    chip2 = demodata.get_kpts_dummy_img(kpts2)
    #kpts2 = ktool.get_grid_kpts()
    fm = demodata.make_dummy_fm(len(kpts1))
    return chip1, chip2, kpts1, kpts2, fm


def get_dummy_test_vars1(fname1='tsukuba_r', fname2='tsukuba_l'):
    import utool as ut
    from vtool_ibeis import image as gtool
    from vtool_ibeis import features as feattool
    fpath1 = ut.grab_test_imgpath(fname1)
    fpath2 = ut.grab_test_imgpath(fname2)
    kpts1, vecs1 = feattool.extract_features(fpath1)
    kpts2, vecs2 = feattool.extract_features(fpath2)
    chip1 = gtool.imread(fpath1)
    chip2 = gtool.imread(fpath2)
    #chip1_shape = vt.gtool.open_image_size(fpath1)
    #chip2_shape = gtool.open_image_size(fpath2)
    #dlen_sqrd2 = chip2_shape[0] ** 2 + chip2_shape[1]
    #testtup = (rchip1, rchip2, kpts1, vecs1, kpts2, vecs2, dlen_sqrd2)
    import vtool_ibeis as vt
    checks = 800
    flann_params = {
        'algorithm': 'kdtree',
        'trees': 8
    }
    #pseudo_max_dist_sqrd = (np.sqrt(2) * 512) ** 2
    pseudo_max_dist_sqrd = 2 * (512 ** 2)
    flann = vt.flann_cache(vecs1, flann_params=flann_params)
    try:
        from vtool_ibeis._pyflann_backend import pyflann
    except Exception:
        import pytest
        pytest.skip()
    try:
        fx2_to_fx1, _fx2_to_dist = flann.nn_index(vecs2, num_neighbors=2, checks=checks)
    except pyflann.FLANNException:
        print('vecs1.shape = %r' % (vecs1.shape,))
        print('vecs2.shape = %r' % (vecs2.shape,))
        print('vecs1.dtype = %r' % (vecs1.dtype,))
        print('vecs2.dtype = %r' % (vecs2.dtype,))
        raise
    fx2_to_dist = np.divide(_fx2_to_dist, pseudo_max_dist_sqrd)
    fx2_to_ratio = np.divide(fx2_to_dist.T[0], fx2_to_dist.T[1])
    ratio_thresh = .625
    fx2_to_isvalid = fx2_to_ratio < ratio_thresh
    fx2_m = np.where(fx2_to_isvalid)[0]
    fx1_m = fx2_to_fx1.T[0].take(fx2_m)
    #fs_RAT = np.subtract(1.0, fx2_to_ratio.take(fx2_m))
    fm_RAT = np.vstack((fx1_m, fx2_m)).T
    fm = fm_RAT
    return chip1, chip2, kpts1, kpts2, fm


def get_stashed_test_vars():
    import utool as ut
    chip1, chip2, kpts1, kpts2, fm, homog_tup, aff_tup = ut.load_testdata(
        'chip1', 'chip2', 'kpts1', 'kpts2', 'fm', 'homog_tup', 'aff_tup')
    return chip1, chip2, kpts1, kpts2, fm


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m tests.test_spatial_verification
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import numpy as np
import utool as ut
import ubelt as ub  # NOQA
import cv2


def bboxes_from_vert_list(verts_list, castint=False):
    """ Fit the bounding polygon inside a rectangle """
    return [bbox_from_verts(verts, castint=castint) for verts in verts_list]


def verts_list_from_bboxes_list(bboxes_list):
    """ Create a four-vertex polygon from the bounding rectangle """
    return [verts_from_bbox(bbox) for bbox in bboxes_list]


def verts_from_bbox(bbox, close=False):
    r"""
    Args:
        bbox (tuple):  bounding box in the format (x, y, w, h)
        close (bool): (default = False)

    Returns:
        list: verts

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> bbox = (10, 10, 50, 50)
        >>> close = False
        >>> verts = verts_from_bbox(bbox, close)
        >>> result = ('verts = %s' % (str(verts),))
        >>> print(result)
        verts = ((10, 10), (60, 10), (60, 60), (10, 60))
    """
    x1, y1, w, h = bbox
    x2 = (x1 + w)
    y2 = (y1 + h)
    if close:
        # Close the verticies list (for drawing lines)
        verts = ((x1, y1), (x2, y1), (x2, y2), (x1, y2), (x1, y1))
    else:
        verts = ((x1, y1), (x2, y1), (x2, y2), (x1, y2))
    return verts


def bbox_from_verts(verts, castint=False):
    x = min(x[0] for x in verts)
    y = min(y[1] for y in verts)
    w = max(x[0] for x in verts) - x
    h = max(y[1] for y in verts) - y
    if castint:
        return (int(x), int(y), int(w), int(h))
    else:
        return (x, y, w, h)


def draw_border(img_in, color=(0, 128, 255), thickness=2, out=None):
    r"""
    Args:
        img_in (ndarray[uint8_t, ndim=2]):  image data
        color (tuple): in bgr
        thickness (int):
        out (None):

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> img_in = vt.imread(ut.grab_test_imgpath('carl'))
        >>> color = (0, 128, 255)
        >>> thickness = 20
        >>> out = None
        >>> # xdoctest: +REQUIRES(module:plottool_ibeis)
        >>> img = draw_border(img_in, color, thickness, out)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(img)
        >>> pt.show_if_requested()
    """
    h, w = img_in.shape[0:2]
    #verts = verts_from_bbox((0, 0, w, h))
    #verts = verts_from_bbox((0, 0, w - 1, h - 1))
    half_thickness = thickness // 2
    verts = verts_from_bbox((half_thickness, half_thickness,
                             w - thickness, h - thickness))
    # FIXME: adjust verts and draw lines here to fill in the corners correctly
    img = draw_verts(img_in, verts, color=color, thickness=thickness, out=out)
    return img


def draw_verts(img_in, verts, color=(0, 128, 255), thickness=2, out=None):
    r"""
    Args:
        img_in (?):
        verts (?):
        color (tuple):
        thickness (int):

    Returns:
        ndarray[uint8_t, ndim=2]: img -  image data

    CommandLine:
        python -m vtool_ibeis.geometry --test-draw_verts --show
        python -m vtool_ibeis.geometry --test-draw_verts:0 --show
        python -m vtool_ibeis.geometry --test-draw_verts:1 --show

    References:
        http://docs.opencv.org/modules/core/doc/drawing_functions.html#line

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> img_in = vt.imread(ut.grab_test_imgpath('carl'))
        >>> verts = ((10, 10), (10, 100), (100, 100), (100, 10))
        >>> color = (0, 128, 255)
        >>> thickness = 2
        >>> # execute function
        >>> out = None
        >>> img = draw_verts(img_in, verts, color, thickness, out)
        >>> assert img_in is not img
        >>> assert out is not img
        >>> assert out is not img_in
        >>> # verify results
        >>> # xdoctest: +REQUIRES(--show)
        >>> pt.imshow(img)
        >>> pt.show_if_requested()

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> img_in = vt.imread(ut.grab_test_imgpath('carl'))
        >>> verts = ((10, 10), (10, 100), (100, 100), (100, 10))
        >>> color = (0, 128, 255)
        >>> thickness = 2
        >>> out = img_in
        >>> # execute function
        >>> img = draw_verts(img_in, verts, color, thickness, out)
        >>> assert img_in is img, 'should be in place'
        >>> assert out is img, 'should be in place'
        >>> # verify results
        >>> # xdoctest: +REQUIRES(--show)
        >>> pt.imshow(img)
        >>> pt.show_if_requested()


    out = img_in = np.zeros((500, 500, 3), dtype=np.uint8)
    """
    if out is None:
        out = np.copy(img_in)
    if isinstance(verts, np.ndarray):
        verts = verts.tolist()
    connect = True
    if connect:
        line_list_sequence = zip(verts[:-1], verts[1:])
        line_tuple_sequence = ((tuple(p1_), tuple(p2_)) for (p1_, p2_) in line_list_sequence)
        cv2.line(out, tuple(verts[0]), tuple(verts[-1]), color, thickness)
        for (p1, p2) in line_tuple_sequence:
            cv2.line(out, p1, p2, color, thickness)
            #print('p1, p2: (%r, %r)' % (p1, p2))
    else:
        for count, p in enumerate(verts, start=1):
            cv2.circle(out, tuple(p), count, color, thickness=1)
    return out


def closest_point_on_line_segment(p, e1, e2):
    """
    Finds the closet point from p on line segment (e1, e2)

    Args:
        p (ndarray): and xy point
        e1 (ndarray): the first xy endpoint of the segment
        e2 (ndarray): the second xy endpoint of the segment

    Returns:
        ndarray: pt_on_seg - the closest xy point on (e1, e2) from p

    References:
        http://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
        http://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment

    CommandLine:
        python -m vtool_ibeis.geometry --exec-closest_point_on_line_segment --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> #bbox = np.array([10, 10, 10, 10], dtype=float)
        >>> #verts_ = np.array(vt.verts_from_bbox(bbox, close=True))
        >>> #R = vt.rotation_around_bbox_mat3x3(vt.TAU / 3, bbox)
        >>> #verts = vt.transform_points_with_homography(R, verts_.T).T
        >>> verts = np.array([[ 21.83012702,  13.16987298],
        >>>                   [ 16.83012702,  21.83012702],
        >>>                   [  8.16987298,  16.83012702],
        >>>                   [ 13.16987298,   8.16987298],
        >>>                   [ 21.83012702,  13.16987298]])
        >>> rng = np.random.RandomState(0)
        >>> p_list = rng.rand(64, 2) * 20 + 5
        >>> close_pts = np.array([closest_point_on_vert_segments(p, verts) for p in p_list])
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.ensureqt()
        >>> pt.plt.plot(p_list.T[0], p_list.T[1], 'ro', label='original point')
        >>> pt.plt.plot(close_pts.T[0], close_pts.T[1], 'rx', label='closest point on shape')
        >>> for x, y in list(zip(p_list, close_pts)):
        >>>     z = np.array(list(zip(x, y)))
        >>>     pt.plt.plot(z[0], z[1], 'r--')
        >>> pt.plt.legend()
        >>> pt.plt.plot(verts.T[0], verts.T[1], 'b-')
        >>> pt.plt.xlim(0, 30)
        >>> pt.plt.ylim(0, 30)
        >>> pt.plt.axis('equal')
        >>> ut.show_if_requested()
    """
    # shift e1 to origin
    de = (dx, dy) = e2 - e1
    # make point vector wrt orgin
    pv = p - e1
    # Project pv onto de
    mag = np.linalg.norm(de)
    pt_on_line_ = pv.dot(de / mag) * de / mag
    # Check if normalized dot product is between 0 and 1
    # Determines if pt is between 0,0 and de
    t = de.dot(pt_on_line_) / mag ** 2
    # t is an interpolation factor indicating how far past the line segment we
    # are. We are on the line segment if it is in the range 0 to 1.
    if t < 0:
        pt_on_seg = e1
    elif t > 1:
        pt_on_seg = e2
    else:
        pt_on_seg = pt_on_line_ + e1
    return pt_on_seg


def distance_to_lineseg(p, e1, e2):
    import vtool_ibeis as vt
    close_pt = vt.closest_point_on_line_segment(p, e1, e2)
    dist_to_lineseg = vt.L2(p, close_pt)
    return dist_to_lineseg


def closest_point_on_line(p, e1, e2):
    """
    e1 and e2 define two points on the line.
    Does not clip to the segment.

    CommandLine:
        python -m vtool_ibeis.geometry closest_point_on_line --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> verts = np.array([[ 21.83012702,  13.16987298],
        >>>                   [ 16.83012702,  21.83012702],
        >>>                   [  8.16987298,  16.83012702],
        >>>                   [ 13.16987298,   8.16987298],
        >>>                   [ 21.83012702,  13.16987298]])
        >>> rng = np.random.RandomState(0)
        >>> p_list = rng.rand(64, 2) * 20 + 5
        >>> close_pts = []
        >>> for p in p_list:
        >>>     candidates = [closest_point_on_line(p, e1, e2) for e1, e2 in ut.itertwo(verts)]
        >>>     dists = np.array([vt.L2_sqrd(p, new_pt) for new_pt in candidates])
        >>>     close_pts.append(candidates[dists.argmin()])
        >>> close_pts = np.array(close_pts)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.ensureqt()
        >>> pt.plt.plot(p_list.T[0], p_list.T[1], 'ro', label='original point')
        >>> pt.plt.plot(close_pts.T[0], close_pts.T[1], 'rx', label='closest point on shape')
        >>> for x, y in list(zip(p_list, close_pts)):
        >>>     z = np.array(list(zip(x, y)))
        >>>     pt.plt.plot(z[0], z[1], 'r--')
        >>> pt.plt.legend()
        >>> pt.plt.plot(verts.T[0], verts.T[1], 'b-')
        >>> pt.plt.xlim(0, 30)
        >>> pt.plt.ylim(0, 30)
        >>> pt.plt.axis('equal')
        >>> ut.show_if_requested()
    """
    # shift e1 to origin
    de = (dx, dy) = e2 - e1
    # make point vector wrt orgin
    pv = p - e1
    # Project pv onto de
    mag = np.linalg.norm(de)
    pt_on_line_ = pv.dot(de / mag) * de / mag
    pt_on_line = pt_on_line_ + e1
    return pt_on_line


def closest_point_on_vert_segments(p, verts):
    import vtool_ibeis as vt
    candidates = [closest_point_on_line_segment(p, e1, e2) for e1, e2 in ut.itertwo(verts)]
    dists = np.array([vt.L2_sqrd(p, new_pt) for new_pt in candidates])
    new_pts = candidates[dists.argmin()]
    return new_pts


def closest_point_on_bbox(p, bbox):
    """

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> p_list = np.array([[19, 7], [7, 14], [14, 11], [8, 7], [23, 21]], dtype=float)
        >>> bbox = np.array([10, 10, 10, 10], dtype=float)
        >>> [closest_point_on_bbox(p, bbox) for p in p_list]
    """
    import vtool_ibeis as vt
    verts = np.array(vt.verts_from_bbox(bbox, close=True))
    new_pts = closest_point_on_vert_segments(p, verts)
    return new_pts


def bbox_from_xywh(xy, wh, xy_rel_pos=[0, 0]):
    """ need to specify xy_rel_pos if xy is not in tl already """
    to_tlx = xy_rel_pos[0] * wh[0]
    to_tly = xy_rel_pos[1] * wh[1]
    tl_x = xy[0] - to_tlx
    tl_y = xy[1] - to_tly
    bbox = [tl_x, tl_y, wh[0], wh[1]]
    return bbox


def extent_from_verts(verts):
    bbox = bbox_from_verts(verts)
    extent = extent_from_bbox(bbox)
    return extent


def union_extents(extents):
    extents = np.array(extents)
    xmin = extents.T[0].min()
    xmax = extents.T[1].max()
    ymin = extents.T[2].min()
    ymax = extents.T[3].max()
    return (xmin, xmax, ymin, ymax)


def extent_from_bbox(bbox):
    """
    Args:
        bbox (ndarray): tl_x, tl_y, w, h

    Returns:
        extent (ndarray): tl_x, br_x, tl_y, br_y

    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/geometry.py extent_from_bbox

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> bbox = [0, 0, 10, 10]
        >>> extent = extent_from_bbox(bbox)
        >>> result = ('extent = %s' % (ub.repr2(extent, nl=0),))
        >>> print(result)
        extent = [0, 10, 0, 10]
    """
    tl_x, tl_y, w, h = bbox
    br_x = tl_x + w
    br_y = tl_y + h
    extent = [tl_x, br_x, tl_y, br_y]
    return extent


#def tlbr_from_bbox(bbox):
def bbox_from_extent(extent):
    """
    Args:
        extent (ndarray): tl_x, br_x, tl_y, br_y

    Returns:
        bbox (ndarray): tl_x, tl_y, w, h

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> extent = [0, 10, 0, 10]
        >>> bbox = bbox_from_extent(extent)
        >>> result = ('bbox = %s' % (ub.repr2(bbox, nl=0),))
        >>> print(result)
        bbox = [0, 0, 10, 10]
    """
    tl_x, br_x, tl_y, br_y = extent
    w = br_x - tl_x
    h = br_y - tl_y
    bbox = [tl_x, tl_y, w, h]
    return bbox


def bbox_from_center_wh(center_xy, wh):
    return bbox_from_xywh(center_xy, wh, xy_rel_pos=[.5, .5])


def bbox_center(bbox):
    (x, y, w, h) = bbox
    centerx = x + (w / 2)
    centery = y + (h / 2)
    return centerx, centery


def get_pointset_extents(pts):
    minx, miny = pts.min(axis=0)
    maxx, maxy = pts.max(axis=0)
    bounds = minx, maxx, miny, maxy
    return bounds


def get_pointset_extent_wh(pts):
    minx, miny = pts.min(axis=0)
    maxx, maxy = pts.max(axis=0)
    extent_w = maxx - minx
    extent_h = maxy - miny
    return extent_w, extent_h


def cvt_bbox_xywh_to_pt1pt2(xywh, sx=1.0, sy=1.0, round_=True):
    """ Converts bbox to thumb format with a scale factor"""
    import vtool_ibeis as vt
    (x1, y1, _w, _h) = xywh
    x2 = (x1 + _w)
    y2 = (y1 + _h)
    if round_:
        pt1 = (vt.iround(x1 * sx), vt.iround(y1 * sy))
        pt2 = (vt.iround(x2 * sx), vt.iround(y2 * sy))
    else:
        pt1 = ((x1 * sx), (y1 * sy))
        pt2 = ((x2 * sx), (y2 * sy))
    return (pt1, pt2)


def scale_bbox(bbox, sx, sy=None):
    if sy is None:
        sy = sx
    from vtool_ibeis import linalg
    centerx, centery = bbox_center(bbox)
    S = linalg.scale_around_mat3x3(sx, sy, centerx, centery)
    verts = np.array(verts_from_bbox(bbox))
    vertsT = linalg.transform_points_with_homography(S, verts.T).T
    bboxT = bbox_from_verts(vertsT)
    return bboxT


def scale_extents(extents, sx, sy=None):
    """
    Args:
        extent (ndarray): tl_x, br_x, tl_y, br_y
    """
    bbox = bbox_from_extent(extents)
    bboxT = scale_bbox(bbox, sx, sy)
    extentsT = extent_from_bbox(bboxT)
    return extentsT


def scaled_verts_from_bbox_gen(bbox_list, theta_list, sx=1, sy=1):
    r"""
    Helps with drawing scaled bbounding boxes on thumbnails

    Args:
        bbox_list (list): bboxes in x,y,w,h format
        theta_list (list): rotation of bounding boxes
        sx (float): x scale factor
        sy (float): y scale factor

    Yeilds:
        new_verts - vertices of scaled bounding box for every input

    CommandLine:
        python -m vtool_ibeis.image --test-scaled_verts_from_bbox_gen

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> # build test data
        >>> bbox_list = [(10, 10, 100, 100)]
        >>> theta_list = [0]
        >>> sx = .5
        >>> sy = .5
        >>> # execute function
        >>> new_verts_list = list(scaled_verts_from_bbox_gen(bbox_list, theta_list, sx, sy))
        >>> result = str(new_verts_list)
        >>> # verify results
        >>> print(result)
        [[[5, 5], [55, 5], [55, 55], [5, 55], [5, 5]]]
    """
    # TODO: input verts support and better name
    for bbox, theta in zip(bbox_list, theta_list):
        new_verts = scaled_verts_from_bbox(bbox, theta, sx, sy)
        yield new_verts


def scaled_verts_from_bbox(bbox, theta, sx, sy):
    """
    Helps with drawing scaled bbounding boxes on thumbnails

    """
    if bbox is None:
        return None
    from vtool_ibeis import linalg
    # Transformation matrixes
    R = linalg.rotation_around_bbox_mat3x3(theta, bbox)
    S = linalg.scale_mat3x3(sx, sy)
    # Get verticies of the annotation polygon
    verts = verts_from_bbox(bbox, close=True)
    # Rotate and transform to thumbnail space
    xyz_pts = linalg.add_homogenous_coordinate(np.array(verts).T)
    trans_pts = linalg.remove_homogenous_coordinate(S.dot(R).dot(xyz_pts))
    new_verts = np.round(trans_pts).astype(np.int32).T.tolist()
    return new_verts


def point_inside_bbox(point, bbox):
    r"""
    Flags points that are strictly inside a bounding box.
    Points on the boundary are not considered inside.

    Args:
        point (ndarray): one or more points to test (2xN)
        bbox (tuple): a bounding box in  (x, y, w, h) format

    Returns:
        bool or ndarray: True if the point is in the bbox

    CommandLine:
        python -m vtool_ibeis.geometry point_inside_bbox --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.geometry import *  # NOQA
        >>> point = np.array([
        >>>     [3, 2], [4, 1], [2, 3], [1, 1], [0, 0],
        >>>     [4, 9.5], [9, 9.5], [7, 2], [7, 8], [9, 3]
        >>> ]).T
        >>> bbox = (3, 2, 5, 7)
        >>> flag = point_inside_bbox(point, bbox)
        >>> flag = flag.astype(int)
        >>> result = ('flag = %s' % (ub.repr2(flag),))
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> verts = np.array(verts_from_bbox(bbox, close=True))
        >>> pt.plot(verts.T[0], verts.T[1], 'b-')
        >>> pt.plot(point[0][flag], point[1][flag], 'go')
        >>> pt.plot(point[0][~flag], point[1][~flag], 'rx')
        >>> pt.plt.xlim(0, 10); pt.plt.ylim(0, 10)
        >>> pt.show_if_requested()
        flag = np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0])
    """
    x, y = point
    tl_x, br_x, tl_y, br_y = extent_from_bbox(bbox)
    inside_x = np.logical_and(tl_x < x, x < br_x)
    inside_y = np.logical_and(tl_y < y, y < br_y)
    flag = np.logical_and(inside_x, inside_y)
    return flag


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.geometry
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
